#!/usr/bin/env python

"""
This module will scan RDKit sources searching for docstrings
that lack parameter definitions, or member functions that
do not have an explicit "self" parameter, and will patch
the C++ sources accordingly.
"""

import sys
import os
import re
import itertools
import glob
import json
import importlib
import queue
import subprocess
import multiprocessing
import shutil
import traceback
import logging
import tempfile
from threading import Thread
from pathlib import Path


RDKIT_MODULE_NAME = "rdkit"
CLANG_CPP_EXE = os.environ.get("CLANG_CPP_EXE", "clang++")
CLANG_FORMAT_EXE = os.environ.get("CLANG_FORMAT_EXE", "clang-format")

CLANG_PYTHON_BINDINGS_PATH = os.environ.get("CLANG_PYTHON_BINDINGS_PATH", None)
if CLANG_PYTHON_BINDINGS_PATH is None:
    raise ValueError(f"Please set CLANG_PYTHON_BINDINGS_PATH to the absolute path to the bindings/python directory under the clang tree")

if CLANG_PYTHON_BINDINGS_PATH not in sys.path:
    sys.path.insert(0, CLANG_PYTHON_BINDINGS_PATH)

if sys.platform.startswith("linux"):
    CLANG_LIBCLANG = "libclang.so"
elif sys.platform.startswith("darwin"):
    CLANG_LIBCLANG = "libclang.dylib"
elif sys.platform.startswith("win32"):
    CLANG_LIBCLANG = "clang.dll"
else:
  raise ValueError(f"Unsupported platform {sys.platform}")
CLANG_LIBRARY_PATH = os.environ.get("CLANG_LIBRARY_PATH", None)

if CLANG_LIBRARY_PATH is None:
    res = subprocess.run([CLANG_CPP_EXE, "-print-resource-dir"], capture_output=True, check=True)
    lib_path = res.stdout.decode("utf-8").strip()
    have_libclang = False
    while lib_path != os.sep:
        libclang_path = os.path.join(lib_path, CLANG_LIBCLANG)
        have_libclang = os.path.exists(libclang_path)
        if have_libclang:
            break
        lib_path = os.path.dirname(lib_path)
    if have_libclang:
        CLANG_LIBRARY_PATH = libclang_path

if CLANG_LIBRARY_PATH is None:
    raise ValueError(f"Please set CLANG_LIBRARY_PATH to the absolute path to {CLANG_LIBCLANG}")

clang_cindex = importlib.import_module("clang.cindex")
clang_cindex.Config.set_library_file(CLANG_LIBRARY_PATH)
TranslationUnit = clang_cindex.TranslationUnit
CursorKind = clang_cindex.CursorKind


class FunctionDef:
    def __init__(self, def_cursor, func_name, is_staticmethod, level):
        self.def_cursor = def_cursor
        self.func_name = func_name
        self.is_staticmethod = is_staticmethod
        self.level = level

class WorkerResult:
    """Result generated by a Worker thread.
    """
    def __init__(self, worker_idx):
        self.worker_idx = worker_idx
        self.processed_cpp_files = set()
        self.proc_error = ""


class DictLike(dict):
    """Base class that confers dict-like behavior
    to any class derived from it.
    """
    def __getattr__(self, key):
        return self[key]

    def __setattr__(self, key, value):
        self[key] = value

    def to_json(self):
        """Serialize class to a JSON string.

        Returns:
            str: JSON-serialized class content
        """
        return json.dumps(self)

    @classmethod
    def from_json(cls, j):
        """Deserialize class from a JSON string.

        Args:
            j (str): JSON string

        Returns:
            cls: an instance of cls
        """
        instance = cls()
        for k, v in json.loads(j).items():
            setattr(instance, k, v)
        return instance


class ClassInfo(DictLike):
    def __init__(self, hash, parents):
        self.hash = hash
        self.parents = parents
        self.python_class_name = None
        self.cpp_class_name = None

class CppFile(DictLike):
    """Class associated to a single C++ file.
    """

    QUOTED_FIELD_REGEX = re.compile(r"\"([^\"]*)\"")
    EXTRACT_BASE_CLASS_NAME_REGEX = re.compile(r"\s*(\S+)\s*<[^>]+>\s*$")
    EXTRACT_INIT_ARGS = re.compile(r"^<(.*)\s>+\s$")
    IS_TEMPLATE_TYPE = re.compile(r"^T\d*$")
    SELF_LITERAL = "self"

    def __init__(self, cpp_path=None):
        """_summary_

        Args:
            cpp_path (str, optional): absolute path to a C++ file.
            Defaults to None.
        """
        self.cpp_path = cpp_path
        self.arg1_func_defs = []
        # type_ref_dict is a dictionary of class alias typedefs
        # relating typedefs to the actual class name
        # found while walking the AST tree
        self.type_ref_dict = {}
        self.ast_error = None
        self.ast_warning = None

    @property
    def ast_path(self):
        """Return absolute file to .ast file associated to this C++ file. 

        Returns:
            str: absolute file to .ast file associated to this C++ file
        """
        return f"{self.cpp_path_noext}.ast"

    @property
    def cpp_path_noext(self):
        """Return absolute path to this C++ file without extension.

        Returns:
            str: absolute path to this C++ file without extension
        """
        return os.path.splitext(self.cpp_path)[0]

    @property
    def cpp_path_ext(self):
        """Return extension of this C++ file.

        Returns:
            str: extension of this C++ file
        """
        return os.path.splitext(self.cpp_path)[1]

    def extract_quoted_content(self, s):
        """Extract string between double quotes.

        Args:
            s (str): string to be parsed

        Returns:
            str: string between quotes, or None
            if there were no quotes.
        """
        res = None
        m = self.QUOTED_FIELD_REGEX.match(s)
        if m:
            res = m.group(1)
        return res

    def generate_ast(self, clang_flags):
        """Generate AST file with clang++.

        Args:
            clang_flags (list): list of flags to be passed to clang++

        Returns:
            bool: True if success, False if failure
        """
        res = False
        try:
            cpp_file = self.cpp_path
            cpp_dir = os.path.dirname(cpp_file)
            self.ast_error = ""
            proc = subprocess.run([CLANG_CPP_EXE] + clang_flags + [cpp_file],
                                  capture_output=True, cwd=cpp_dir)
            if proc.returncode:
                self.ast_error += f"{cpp_file}: Failed to parse with clang."
                if proc.stderr:
                    self.ast_error += "\nError was:\n" + proc.stderr.decode("utf-8")
            else:
                res = True 
                if proc.stderr:
                    self.ast_warning = proc.stderr.decode("utf-8")
        except Exception as e:
            res = False
            tb = traceback.format_exc()
            self.ast_error += f"{cpp_file}: Failed to run clang\n{str(e)}\n{str(tb)}\n"
        return res

    def recurse_ast_cursor(self, cursor, hnd=sys.stdout, recursion_level=0):
        """Recursively walk on the AST tree and write a dump to hnd.

        Args:
            cursor (Cursor): current cursor position
            hnd (file-like object, optional): Handle to which the dump should be written.
            Defaults to sys.stdout.
            recursion_level (int, optional): Used to indent the dump. Defaults to 0.
        """
        recursion_level += 1
        tabs = "".join("  " for i in range(recursion_level))
        for child in cursor.get_children():
            if child.kind == CursorKind.TYPEDEF_DECL:
                typedef_key = child.spelling
                typedef_value = None
                for child2 in child.get_children():
                    if child2.kind in (CursorKind.TYPE_REF, CursorKind.TEMPLATE_REF):
                        typedef_value = child2.spelling
                        break
                if typedef_value is not None:
                    typedef_value = typedef_value.split(" ")[-1]
                    self.type_ref_dict[typedef_key] = typedef_value
            print(f"{tabs}{child.kind}:{child.spelling}", file=hnd)
            hnd.flush()
            self.recurse_ast_cursor(child, hnd, recursion_level)

    def get_func_name_if_has_arg1_param_r(self, cursor, arg1_func_names):
        """Return name of function that needs fixing.

        Args:
            cursor (Cursor): current cursor position
            arg1_func_names (iterable): function names that need fixing

        Returns:
            str: function name
        """
        if cursor.kind == CursorKind.STRING_LITERAL:
            func_name = self.extract_quoted_content(cursor.spelling)
            if func_name is not None and func_name in arg1_func_names:
                return func_name
        for child in cursor.get_children():
            res = self.get_func_name_if_has_arg1_param_r(child, arg1_func_names)
            if res is not None:
                return res
        return None

    def have_decl_ref_expr_r(self, cursor):
        """Check if there is a "def" child below this cursor.

        Args:
            cursor (Cursor): current cursor position

        Returns:
            bool: True if there is a "def" child
        """
        if cursor.kind == CursorKind.DECL_REF_EXPR and cursor.spelling == "def":
            return True
        for child in cursor.get_children():
            if self.have_decl_ref_expr_r(child):
                return True
        return False

    def find_non_class_defs(self, cursor, class_method_node_hashes, arg1_func_names):
        """Recursively find free function nodes that need fixing.

        Args:
            cursor (Cursor): current cursor position
            non_class_defs (dict): dict relating a function name to a list of nodes
            class_method_node_hashes (iterable): set of hashes corresponding to class method
            nodes that need fixing previously found
            arg1_func_names (iterable): function names that need fixing
        """
        non_class_defs = {}
        self.find_non_class_defs_r(cursor, non_class_defs, class_method_node_hashes, set(arg1_func_names))
        return non_class_defs

    def find_non_class_defs_r(self, cursor, non_class_defs, class_method_node_hashes, arg1_func_names):
        """Find free function nodes that need fixing (recursive).

        Args:
            cursor (Cursor): current cursor position
            non_class_defs (dict): dict relating a function name to a list of nodes
            class_method_node_hashes (iterable): set of hashes corresponding to class method
            nodes that need fixing previously found
            arg1_func_names (iterable): function names that need fixing
        """
        if (cursor.kind == CursorKind.CALL_EXPR and cursor.spelling == "def"
            and self.have_decl_ref_expr_r(cursor) and cursor.hash not in class_method_node_hashes):
            func_name = self.get_func_name_if_has_arg1_param_r(cursor, arg1_func_names)
            if func_name is not None:
                node_list = non_class_defs.get(func_name, [])
                node_list.append(cursor)
                non_class_defs[func_name] = node_list
        for child in cursor.get_children():
            self.find_non_class_defs_r(child, non_class_defs, class_method_node_hashes, arg1_func_names)

    def find_nodes(self, cursor):
        """Recursively walk on the AST tree and associate node
        to their python::class_ hash.

        Args:
            cursor (Cursor): current cursor position

        Returns:
            dict[int, ClassInfo]: dict associating a class hash
            to a ClassInfo instance
        """
        class_info_by_class_hash = {}
        self.find_nodes_r(cursor, [], class_info_by_class_hash)
        return class_info_by_class_hash

    def find_nodes_r(self, cursor, parents, class_info_by_class_hash):
        """Walk on the AST tree and associate node
        to their python::class_ hash (recursive)

        Args:
            cursor (Cursor): current cursor position
            parents (list): growing list of parents
            class_info_by_class_hash (dict): dict associating a class hash
            to a ClassInfo instance
        """
        if cursor.kind == CursorKind.CALL_EXPR:
            parents = [cursor] + parents
        if cursor.kind in (CursorKind.CALL_EXPR, CursorKind.TEMPLATE_REF) and cursor.spelling == "class_":
            class_hash = cursor.hash
            prev_class_info = class_info_by_class_hash.get(class_hash, None)
            if prev_class_info is None or len(parents) + 1 > len(prev_class_info.parents):
                class_info = ClassInfo(class_hash, parents)
                class_info_by_class_hash[class_hash] = class_info
                for child in cursor.get_children():
                    if child.kind == CursorKind.TYPE_REF:
                        cpp_class_name = child.spelling.split(" ")[-1].split("::")[-1]
                        class_info.cpp_class_name = self.type_ref_dict.get(cpp_class_name, cpp_class_name)
                        break
                return
        for child in cursor.get_children():
            self.find_nodes_r(child, parents, class_info_by_class_hash)

    def find_class_name_r(self, class_info, cursor, found_class_names, arg1_func_byclass_dict):
        """Find the name of the python::class_ associated to class_hash.

        Args:
            class_info (ClassInfo): ClassInfo instance
            cursor (Cursor): current cursor position
            found_class_names (dict): dict relating class hash to class name
            arg1_func_byclass_dict (dict): dict relating class names to methods
            that need fixing. Also free functions are included under class name
            FixSignatures.NO_CLASS_KEY

        Returns:
            bool: True if the class name corresponding to class_hash was found
        """
        class_hash = class_info.hash
        if cursor.kind == CursorKind.STRING_LITERAL:
            class_name = self.extract_quoted_content(cursor.spelling)
            if (class_name is not None and class_name in arg1_func_byclass_dict
                and class_hash not in found_class_names and class_name not in found_class_names.values()):
                found_class_names[class_hash] = class_name
                class_info.python_class_name = class_name
                return True
        for child in cursor.get_children():
            if self.find_class_name_r(class_info, child, found_class_names, arg1_func_byclass_dict):
                return True
        return False

    def prune_nodes(self, tu_cursor, class_info_by_class_hash, arg1_func_byclass_dict):
        """Return a dict relating class name to a ClassInfo instance.

        Args:
            class_info_by_class_hash (dict): dict relating class hash to class_info
            arg1_func_byclass_dict (dict): dict relating class names to methods

        Returns:
            dict: dict relating class name to a ClassInfo instance
        """
        # populate found_class_names dictionary {class_hash: class_name}
        # with classes that have methods we need to fix
        found_class_names = {}
        for class_hash, class_info in class_info_by_class_hash.items():
            call_expr_class_node = class_info.parents[0]
            # we might not find the class name as STRING_LITERAL for template classes
            self.find_class_name_r(class_info, call_expr_class_node, found_class_names, arg1_func_byclass_dict)
        for class_hash, class_info in class_info_by_class_hash.items():
            # 2nd pass over the whole translation unit to find template classes
            self.find_class_name_r(class_info, tu_cursor, found_class_names, arg1_func_byclass_dict)
        # prune class_hash entries that do not have methods we need to fix
        class_info_by_class_name = {}
        for class_hash in tuple(class_info_by_class_hash.keys()):
            if class_hash in found_class_names.keys():
                found_class_name = found_class_names[class_hash]
                class_info_by_class_name[found_class_name] = class_info_by_class_hash[class_hash]
        return class_info_by_class_name

    def have_python_range_r(self, cursor, requested_level, level=0):
        """Return True if there is a python::range among the children of cursor.

        Args:
            cursor (Cursor): current cursor position
            requested_level (int): requested nesting level
            level (int, optional): current nesting level. Defaults to 0.

        Returns:
            bool: True if there is a python::range among the children of cursor
        """
        level += 1
        for child in cursor.get_children():
            if level == requested_level and child.kind == CursorKind.CALL_EXPR and child.spelling == "range":
                return True
            if self.have_python_range_r(child, requested_level, level):
                return True
        return False

    def find_func_name_r(self, cursor, def_cursor, func_names, func_name_to_hash, def_init_nodes, level=0):
        """Find the Python function name connected to this cursor.

        Args:
            cursor (Cursor): current cursor position
            def_cursor (Cursor): node corrisponding to the function "def"
            func_names (list): function names that need fixing
            func_name_to_hash (dict): dict relating function names to the
            "def" node hash
            def_init_nodes (dict): dict relating node hash to a FunctionDef instance
            level (int, optional): _description_. Defaults to 0.
        """
        log_path = self.cpp_path_noext + ".log"
        is_staticmethod = (def_cursor.spelling == "staticmethod" and level == 1)
        level += 1
        for child in cursor.get_children():
            if child.kind == CursorKind.STRING_LITERAL:
                func_name = self.extract_quoted_content(child.spelling)
                if (func_name is not None and func_name not in ("__init__", "__exit__", "__enter__")
                    and not (func_name == "__iter__" and self.have_python_range_r(def_cursor, level))):
                    try:
                        func_name_idx = func_names.index(func_name)
                        if def_cursor.hash in def_init_nodes:
                            with open(log_path, "a") as hnd:
                                print(f"1) find_func_name_r def_cursor.hash {def_cursor.hash} level {level} func_name {func_name} kind {def_cursor.kind} tokens {[t.spelling for t in def_cursor.get_tokens()]}", file=hnd)
                                hnd.flush()
                            assert def_cursor.hash not in def_init_nodes
                        func_names.pop(func_name_idx)
                        func_name_to_hash[func_name] = def_cursor.hash
                        def_init_nodes[def_cursor.hash] = FunctionDef(def_cursor, func_name, is_staticmethod, level)
                        with open(log_path, "a") as hnd:
                            print(f"2) find_func_name_r def_cursor.hash {def_cursor.hash} level {level} func_name {func_name} kind {def_cursor.kind} tokens {[t.spelling for t in def_cursor.get_tokens()]}", file=hnd)
                            hnd.flush()
                    except ValueError:
                        hash_for_func_name = func_name_to_hash.get(func_name, None)
                        if hash_for_func_name is not None and hash_for_func_name != def_cursor.hash:
                            prev_function_def = def_init_nodes.get(hash_for_func_name, None)
                            if prev_function_def is not None:
                                if prev_function_def.is_staticmethod and not is_staticmethod:
                                    def_init_nodes[hash_for_func_name] = FunctionDef(def_cursor, func_name, True, level)
                                elif not prev_function_def.is_staticmethod and is_staticmethod:
                                    def_init_nodes[hash_for_func_name] = FunctionDef(prev_function_def.def_cursor, func_name, True, prev_function_def.level)
                                elif (not (prev_function_def.is_staticmethod ^ is_staticmethod)
                                      and prev_function_def.func_name == func_name
                                      and def_cursor.kind != CursorKind.MEMBER_REF_EXPR
                                      and def_cursor.hash not in def_init_nodes
                                      and ([t.spelling for t in def_cursor.get_tokens()].count("def")
                                           != [t.spelling for t in prev_function_def.def_cursor.get_tokens()].count("def"))
                                      and self.is_last_def(func_name, list(def_cursor.get_tokens()))):
                                    with open(log_path, "a") as hnd:
                                        print(f"3) find_func_name_r def_cursor.hash {def_cursor.hash} level {level} func_name {func_name} kind {def_cursor.kind} tokens {[t.spelling for t in def_cursor.get_tokens()]}", file=hnd)
                                        hnd.flush()
                                    def_init_nodes[def_cursor.hash] = FunctionDef(def_cursor, func_name, is_staticmethod, prev_function_def.level)
                elif func_name == "__init__":
                    def_init_nodes[cursor.hash] = FunctionDef(cursor, "__init__", False, level)
            self.find_func_name_r(child, def_cursor, func_names, func_name_to_hash, def_init_nodes, level)

    def find_cpp_func_r(self, cursor, requested_level, func_name, param_count=-1, level=0):
        """Find the C++ function corresponding to this cursor.

        Args:
            cursor (Cursor): current cursor position
            requested_level (int): requested nesting level
            param_count (int, optional): parameter count for this function. Defaults to -1.
            level (int, optional): current nesting level. Defaults to 0.

        Returns:
            tuple: (param_count, func_name) tuple. func_name can be prefixed with the
            class name if the function is a class method.
        """
        log_path = self.cpp_path_noext + ".log"
        res = None
        level += 1
        for child in cursor.get_children():
            if level == requested_level - 1 and "CAST" in str(child.kind):
                param_count = 0
            if level == requested_level:
                if child.kind == CursorKind.DECL_REF_EXPR and child.spelling != "def":
                    res = child
                elif child.kind == CursorKind.PARM_DECL:
                    assert param_count != -1
                    param_count += 1
                elif child.kind in (CursorKind.UNARY_OPERATOR, CursorKind.UNEXPOSED_EXPR):
                    for child2 in child.get_children():
                        if child2.kind == CursorKind.DECL_REF_EXPR:
                            res = child2
                            break
                elif child.kind == CursorKind.CALL_EXPR and child.spelling == "make_constructor":
                    for child2 in child.get_children():
                        if child2.kind in (CursorKind.UNARY_OPERATOR, CursorKind.UNEXPOSED_EXPR) and not child2.spelling:
                            for child3 in child2.get_children():
                                if child3.kind == CursorKind.DECL_REF_EXPR:
                                    res = child3
                                    break
                            if res is not None:
                                break
            if res is None:
                res = self.find_cpp_func_r(child, requested_level, func_name, param_count, level)
        if res is not None and not isinstance(res, tuple):
            decl_ref = res
            res = None
            for child in decl_ref.get_children():
                if child.kind == CursorKind.TEMPLATE_REF and decl_ref.spelling:
                    template_ref = child.spelling.split("::")[-1]
                    with open(log_path, "a") as hnd:
                        print(f"1) find_cpp_func_r template_ref {template_ref}", file=hnd)
                        hnd.flush()
                    template_ref = self.type_ref_dict.get(template_ref, template_ref)
                    with open(log_path, "a") as hnd:
                        print(f"2) find_cpp_func_r template_ref {template_ref}", file=hnd)
                        hnd.flush()
                    res = template_ref + "::" + decl_ref.spelling
                    break
                elif child.kind == CursorKind.TYPE_REF and decl_ref.spelling:
                    type_ref = child.spelling.split("::")[-1]
                    with open(log_path, "a") as hnd:
                        print(f"3) find_cpp_func_r type_ref {type_ref}", file=hnd)
                        hnd.flush()
                    type_ref = self.type_ref_dict.get(type_ref, type_ref)
                    with open(log_path, "a") as hnd:
                        print(f"4) find_cpp_func_r type_ref {type_ref}", file=hnd)
                        hnd.flush()
                    res = type_ref + "::" + decl_ref.spelling
                    break
                elif child.kind == CursorKind.OVERLOADED_DECL_REF and not decl_ref.spelling and child.spelling:
                    decl_ref = child
            if res is None and decl_ref.spelling:
                res = decl_ref.spelling
            res = (param_count, res)
        return res

    def extract_base_class_name(self, cpp_class_name):
        """Extract the base class name, if present.

        Args:
            cpp_class_name (str): C++ class name

        Returns:
            str: base class name, if present, otherwise input class name
        """
        m = self.EXTRACT_BASE_CLASS_NAME_REGEX.match(cpp_class_name)
        if m:
            cpp_class_name = m.group(1)
        return cpp_class_name.split("::")[-1]

    def find_cpp_class_r(self, cursor, cpp_class_name, func_name):
        """Find the node corresponding to the func_name method of the
        C++ class cpp_class_name.
        If the node cannot be found, it returns the base class name
        instead, so a new recursive search on the base class can
        be carried out.

        Args:
            cursor (Cursor): current cursor position
            cpp_class_name (str): C++ class name
            func_name (str): C++ method name

        Returns:
            cursor|str: node corresponding to the method, or
            base class name if the method could not be found
        """
        res = None
        base_cpp_class_name = None
        for child in cursor.get_children():
            if (child.kind in (CursorKind.CLASS_DECL, CursorKind.CLASS_TEMPLATE, CursorKind.STRUCT_DECL)
                and child.spelling == cpp_class_name):
                for child2 in child.get_children():
                    if child2.spelling.split("<")[0] == func_name and (
                        (child2.kind in (CursorKind.CXX_METHOD, CursorKind.FUNCTION_TEMPLATE)
                         or (func_name == cpp_class_name and child2.kind == CursorKind.CONSTRUCTOR))
                    ):
                        res = child
                        break
                    elif child2.kind == CursorKind.CXX_BASE_SPECIFIER:
                        base_cpp_class_name = self.extract_base_class_name(child2.spelling)
                if res is not None:
                    break
            res = self.find_cpp_class_r(child, cpp_class_name, func_name)
            if res is not None:
                break
        if res is None and base_cpp_class_name is not None:
            return base_cpp_class_name
        return res

    @staticmethod
    def have_param(param_list, param):
        """If param is part of param_list return True and pop it from param_list.

        Args:
            param_list (list[str]): list of parameters
            param (str): parameter

        Returns:
            bool: True if param is part of param_list, False if not
        """
        res = param in param_list
        if res:
            param_list.pop(param_list.index(param))
        return res

    def num_matching_parameters(self, expected_params, params):
        """Find the number of matching params between params
        (list of individual parameter typenames) and expected_params
        (concatenated string of expected parameter typenames)
        Args:
            expected_params (str): concatenated string of expected parameter typenames
            params (list[str]): list of individual parameter typenames

        Returns:
            tuple[int, int]: number of matching params, number of non-matching params
        """
        expected_params_tok = [p.split("::")[-1] for p in expected_params.split()]
        params_tok = [p.split("::")[-1] for p in " ".join(params).split()]
        num_matched_params = [self.have_param(expected_params_tok, p) for p in params_tok].count(True)
        num_non_matched_params = len(params_tok) - num_matched_params
        return num_matched_params, -num_non_matched_params

    def find_cpp_func_params(self, cursor, is_staticmethod, cpp_class_name, func_name,
                             expected_cpp_params, expected_param_count):
        """Find parameter names of a C++ method.
        First we try to find the exact number of expected parameters.
        If we fail, we will accept an overload with a number of parameters
        greater than the expected one, in the assumption that some parameters
        can be optional.

        Args:
            cursor (Cursor): current cursor position
            cpp_class_name (str): C++ class name
            func_name (str): C++ method name
            expected_cpp_params: expected parameter string based
            on the Python function signature. This is a cumulative, concatenated
            string with no spaces which is used when there are multiple overloads
            with the same number of parameters to try and pick the C++ funciton whose
            parameter types best fit the Python signature.
            expected_param_count (int): expected parameter count based
            on the Python function signature

        Returns:
            list[str]: list of parameter names
        """
        self.params = None
        assigned_overloads = None
        if cpp_class_name == func_name:
            key = f"{cpp_class_name}::{cpp_class_name}"
            assigned_overloads = self.assigned_overloads.get(key, [])
            if not assigned_overloads:
                self.assigned_overloads[key] = assigned_overloads
        self.assigned_overloads_for_func = assigned_overloads
        for accept_params_no_type in (False, True):
            self.accept_params_no_type = accept_params_no_type
            for cmp_func in (int.__eq__, int.__gt__):
                self.find_cpp_func_params_r(cursor, cpp_class_name, func_name,
                                            expected_cpp_params, expected_param_count, cmp_func)
                if self.params is not None:
                    break
            if self.params is not None:
                if assigned_overloads is not None and not self.has_template_type(self.params):
                    assigned_overloads.append(self.get_params_hash(self.params))
                break
        if self.params is None:
            params = [f"arg{i + 1}" for i in range(expected_param_count)]
            if not is_staticmethod:
                params.insert(0, "self")
            return params
        return [p for p, _ in self.params]

    def has_template_type(self, params):
        """Find if any parameter in params is of template type.

        Args:
            params (list[tuple[str, str]]): list of (name, type) tuples

        Returns:
            bool: True if params contain parameters of template type
            (i.e., T, optionally followed by a number)
        """
        return any(self.IS_TEMPLATE_TYPE.match(t) for _, t in params)

    @staticmethod
    def get_params_hash(params):
        """Get a hash from function parameters.

        Args:
            params (list[tuple[str, str]]): list of function parameters
            as (parameter name, paramater type) tuples

        Returns:
            tuple: a sorted tuple that can be used as a hash
        """
        return tuple(sorted(params))

    def find_cpp_func_params_r(self, cursor, cpp_class_name, func_name,
                               expected_cpp_params, expected_param_count, cmp_func):
        """Find parameter names of a C++ method (recursive).

        Args:
            cursor (Cursor): current cursor position
            cpp_class_name (str): C++ class name
            func_name (str): C++ method name
            expected_cpp_params: expected parameter string based
            on the Python function signature. This is a cumulative, concatenated
            string with no spaces which is used when there are multiple overloads
            with the same number of parameters to try and pick the C++ funciton whose
            parameter types best fit the Python signature.
            expected_param_count (int): expected parameter count based
            on the Python function signature
            cmp_func (function): the comparator to use between the expected number
            of parameters and the best-fitting found number of parameters
        """
        accepted_kinds = [CursorKind.FUNCTION_DECL,
                          CursorKind.FUNCTION_TEMPLATE]
        if cpp_class_name is not None:
            if func_name != cpp_class_name:
                accepted_kinds.append(CursorKind.CXX_METHOD)
            else:
                accepted_kinds.append(CursorKind.CONSTRUCTOR)
        for child in cursor.get_children():
            if child.kind in accepted_kinds and child.spelling.split("<")[0] == func_name:
                params = [(child2.spelling, " ".join(child3.spelling for child3 in child2.get_children()
                                                     if child3.kind in (CursorKind.TEMPLATE_REF, CursorKind.TYPE_REF)))
                                                     for child2 in child.get_children() if child2.kind == CursorKind.PARM_DECL]
                # certain C++ headers have only the type declaration but no variable name,
                # in that case we replace "" with a dummy parameter name since python::args("")
                # is not acceptable
                params = [(p or f"arg{i + 1}", t) for i, (p, t) in enumerate(params)]
                params_hash = self.get_params_hash(params)
                if self.assigned_overloads_for_func is not None and params_hash in self.assigned_overloads_for_func:
                    continue
                if ((expected_param_count == -1 or cmp_func(len(params), expected_param_count))
                    and (not expected_cpp_params or (self.accept_params_no_type and self.params is None)
                         or (self.params is not None and
                             self.num_matching_parameters(expected_cpp_params, [t for _, t in params])
                             > self.num_matching_parameters(expected_cpp_params, [t for _, t in self.params])))):
                    if expected_param_count != -1:
                        params = params[:expected_param_count]
                    self.params = params
            else:
                self.find_cpp_func_params_r(child, cpp_class_name, func_name, expected_cpp_params, expected_param_count, cmp_func)

    def find_def_init_nodes_in_class_r(self, cursor, func_names, func_name_to_hash, def_init_nodes):
        """Find nodes corresponding to Python constructors and methods for a class.

        Args:
            cursor (Cursor): current cursor position
            func_names (list): function names that need fixing
            func_name_to_hash (dict): dict relating function names to the
            "def" node hash
            def_init_nodes (dict): dict relating node hash to a FunctionDef instance
        """
        if cursor.kind in (CursorKind.CALL_EXPR, CursorKind.TEMPLATE_REF, CursorKind.MEMBER_REF_EXPR):
            if cursor.spelling == "init":
                if cursor.hash not in def_init_nodes:
                    def_init_nodes[cursor.hash] = FunctionDef(cursor, "__init__", False, 0)
            # templated python::class_ may have no "def", so we accept empty spelling
            elif not cursor.spelling or cursor.spelling == "def" or cursor.spelling == "staticmethod":
                self.find_func_name_r(cursor, cursor, func_names, func_name_to_hash, def_init_nodes)
        for child in cursor.get_children():
            self.find_def_init_nodes_in_class_r(child, func_names, func_name_to_hash, def_init_nodes)

    def is_class_hash_among_node_children_r(self, class_hash, node):
        """Return True if class_hash is found among the children of node.

        Args:
            class_hash (int): class hash
            node (cursor): cursor

        Returns:
            bool: True if class_hash is found among the children of node.
        """
        if node.hash == class_hash:
            return True
        for child in node.get_children():
            if self.is_class_hash_among_node_children_r(class_hash, child):
                return True
        return False

    def find_def_init_nodes(self, class_info_by_class_name, arg1_func_byclass_dict):
        """Find Python constructors and methods.

        Args:
            class_info_by_class_name (dict): dict relating class name
            to a ClassInfo instance
            arg1_func_byclass_dict (dict): dict relating class name to methods

        Returns:
            dict: dict relating class name to a (def_init_nodes, func_names) tuple,
            where def_init_nodes is an iterable of constructors and methods,
            and func_names a list of function names that need fixing but could
            not be associated to any methods (currently unused)
        """
        res = {}
        for class_name, class_info in class_info_by_class_name.items():
            def_init_nodes = {}
            func_name_to_hash = {}
            func_names = arg1_func_byclass_dict[class_name]
            for i, node in enumerate(class_info.parents):
                if i and not self.is_class_hash_among_node_children_r(class_info.hash, node):
                    break
                self.find_def_init_nodes_in_class_r(node, func_names, func_name_to_hash, def_init_nodes)
            res[class_name] = (def_init_nodes.values(), func_names)
        return res

    def find_python_args(self, tokens):
        """Insert the "self" arg into existing python::args.

        Args:
            tokens (iterable): iterable of Token objects

        Returns:
            list(tuple)|None: list of tuples with source line number, source column number
            and string to be inserted on that line at that column position
        """
        for i, t in enumerate(tokens):
            if (t.spelling == "python"
                and len(tokens[i:]) > 4
                and tokens[i+1].spelling == "::"
                and tokens[i+2].spelling == "args"
                and tokens[i+3].spelling == "("):
                concat_tokens = "".join(t.spelling for t in tokens)
                # if there are multiple python::args keywords or default parameters,
                # treat them as if they were python::arg
                potential_non_self_token = tokens[i+4]
                arg_name = self.extract_quoted_content(potential_non_self_token.spelling)
                if arg_name != self.SELF_LITERAL and (concat_tokens.count("python::args") > 1 or "=" in concat_tokens):
                    return self.find_python_arg(tokens, "args")
                if arg_name is not None:
                    if arg_name == self.SELF_LITERAL:
                        return []
                    source_loc = potential_non_self_token.extent.start
                    source_line = source_loc.line
                    source_col = source_loc.column
                    return [(source_line, source_col, f"\"{self.SELF_LITERAL}\", ")]
        return None

    def find_python_arg(self, tokens, arg_keyword="arg"):
        """Insert the "self" arg into existing python::arg.

        Args:
            tokens (iterable): iterable of Token objects

        Returns:
            list(tuple)|None: list of tuples with source line number, source column number
            and string to be inserted on that line at that column position
        """
        need_additional_bracket = False
        open_bracket = ""
        for i, t in enumerate(tokens):
            if (t.spelling == "python"
                and len(tokens[i:]) > 4
                and tokens[i+1].spelling == "::"
                and tokens[i+2].spelling == arg_keyword
                and tokens[i+3].spelling == "("):
                bracket_count = 0
                j = i
                while j:
                    j -= 1
                    if tokens[j].spelling == "(":
                        bracket_count += 1
                    elif bracket_count:
                        break
                assert bracket_count
                if bracket_count == 1:
                    need_additional_bracket = True
                    open_bracket = "("
                j = i + 4
                potential_non_self_token = tokens[j]
                arg_name = self.extract_quoted_content(potential_non_self_token.spelling)
                if arg_name is not None:
                    if arg_name == self.SELF_LITERAL:
                        return []
                    source_loc = t.extent.start
                    source_line = source_loc.line
                    source_col = source_loc.column
                    res = [(source_line, source_col, f"{open_bracket}python::{arg_keyword}(\"{self.SELF_LITERAL}\"), ")]
                    if need_additional_bracket:
                        found = False
                        j += 1
                        while tokens[j+1:] and not found:
                            j += 1
                            found = tokens[j].spelling in (",", ")")
                        assert found
                        source_loc = tokens[j].extent.start
                        source_line = source_loc.line
                        source_col = source_loc.column
                        res += [(source_line, source_col, ")")]
                    return res
        return None

    def find_no_arg(self, is_init, tokens, is_staticmethod, cpp_func_name, expected_param_count, cursor, class_info):
        """Insert the appropriate python::args where needed based on the C++
        method parameter names.

        Args:
            is_init (bool): True if the method is a constructor
            tokens (iterable): iterable of Token objects
            is_staticmethod (bool): True if the method is static
            cpp_func_name (str): name of the C++ function
            expected_param_count (int): expected number of parameters
            based on the Python function signature
            cursor (Cursor): current cursor position
            class_info (ClassInfo): ClassInfo instance

        Raises:
            IndexError: in case there are unexpected inconsistencies
            (should never happen)

        Returns:
            list(tuple)|None: list of tuples with source line number, source column number
            and string to be inserted on that line at that column position
        """
        log_path = self.cpp_path_noext + ".log"
        bracket_count = 0
        init_args = ""
        expected_cpp_params = None
        for i, t in enumerate(tokens):
            num_downstream_tokens = len(tokens[i:])
            if is_init:
                open_bracket_count = t.spelling.count("<")
                closed_bracket_count = t.spelling.count(">")
                if open_bracket_count or bracket_count:
                    init_args += t.spelling + " "
                    bracket_count += (open_bracket_count - closed_bracket_count)
                    if bracket_count == 0:
                        if init_args:
                            m = self.EXTRACT_INIT_ARGS.match(init_args)
                            if not m or "python::optional" in init_args:
                                init_args = ""
                                is_init = False
                            else:
                                init_args = m.group(1).replace("<", "").strip()
                        if init_args:
                            cpp_func_name = f"{class_info.cpp_class_name}::{class_info.cpp_class_name}"
                            expected_param_count = 1 + init_args.count(",")
                            expected_cpp_params = init_args
                            init_args = ""
            is_def = (t.spelling == "def")
            if (num_downstream_tokens > 2 and (is_init or is_def)
                and tokens[i+1].spelling == "("):
                need_comma = (tokens[i+2].spelling != ")")
                is_make_constructor = "make_constructor" in (t.spelling for t in tokens)
                python_args = "python::args("
                need_self = not is_staticmethod and not is_make_constructor
                if need_self:
                    python_args += f"\"{self.SELF_LITERAL}\"" + init_args
                if cpp_func_name is not None and expected_param_count is not None and cursor is not None:
                    cpp_func_name_tokens = cpp_func_name.split("::")
                    cpp_class_name = cpp_func_name_tokens[-2] if len(cpp_func_name_tokens) > 1 else None
                    func_name = cpp_func_name_tokens[-1]
                    with open(log_path, "a") as hnd:
                        print(f"1) find_no_arg cpp_func_name {cpp_func_name} cpp_class_name {cpp_class_name} func_name {func_name} expected_param_count {expected_param_count} is_staticmethod {is_staticmethod} tokens {[t.spelling for t in tokens]}", file=hnd)
                        hnd.flush()
                    rename_first_param = need_self
                    if cpp_class_name is not None:
                        while 1:
                            res = self.find_cpp_class_r(cursor, cpp_class_name, func_name)
                            with open(log_path, "a") as hnd:
                                print(f"2) find_no_arg res {res}", file=hnd)
                                hnd.flush()
                            if not isinstance(res, str):
                                break
                            cpp_class_name = res
                        if res is not None:
                            rename_first_param = False
                            cursor = res
                    params = self.find_cpp_func_params(cursor, is_staticmethod, cpp_class_name, func_name, expected_cpp_params, expected_param_count)
                    if rename_first_param:
                        if not params:
                            raise IndexError(f"Expected at least one parameter on {func_name}, found none")
                        params[0] = self.SELF_LITERAL
                    with open(log_path, "a") as hnd:
                        print(f"3) find_no_arg params {params}", file=hnd)
                        hnd.flush()
                    if params is not None:
                        params = ", ".join(f"\"{p}\"" for p in params if p != self.SELF_LITERAL)
                        if params:
                            if need_self:
                                python_args += ", "
                            python_args += params
                python_args += ")"
                if is_init:
                    token_idx = i + 2
                    last_seen_idx = token_idx
                    if need_comma:
                        python_args += ", "
                else:
                    token_idx = i + 4
                    bracket_count = 0
                    last_seen_idx = None
                    while token_idx < len(tokens):
                        s = tokens[token_idx].spelling
                        if (s == "."
                            and token_idx + 1 < len(tokens)
                            and tokens[token_idx + 1].spelling in ("def", "def_pickle", "staticmethod")):
                            break
                        if s and s[0] in ("(", "<"):
                            incr = s.count(s[0])
                            bracket_count += incr
                            last_seen_idx = None
                        elif s and s[0] in (")", ">") and bracket_count:
                            incr = s.count(s[0])
                            assert bracket_count >= incr
                            bracket_count -= incr
                        elif last_seen_idx is None and not bracket_count and s in (",", ")"):
                            last_seen_idx = token_idx
                        token_idx += 1
                    if last_seen_idx is None:
                        raise IndexError(f"Failed to find end of definitions; tokens[i+4]: {[t.spelling for t in tokens[i+4:]]}")
                    if need_comma:
                        python_args = ", " + python_args
                potential_non_self_token = tokens[last_seen_idx]
                source_loc = potential_non_self_token.extent.start
                source_line = source_loc.line
                source_col = source_loc.column
                return [(source_line, source_col, python_args)]
        return None

    def find_func_def(self, func_name, tokens):
        """Find the tokens corresponding to the Python def
        for func_name.

        Args:
            func_name (str): Python function name
            tokens (iterable): iterable of Token objects

        Returns:
            iterable: iterable of Token objects
        """
        for i, t in reversed(list(enumerate(tokens))):
            if (t.spelling == "def"
                and i + 2 < len(tokens)
                and tokens[i+1].spelling == "("
                and tokens[i+2].spelling == f"\"{func_name}\""):
                return tokens[i:]
        return None

    def is_last_def(self, func_name, tokens):
        """Return true if the last "def" in tokens corresponds to func_name.

        Args:
            func_name (str): Python function name
            tokens (iterable): iterable of Token objects

        Returns:
            True if the last "def" in tokens corresponds to func_name, False if not
        """
        for i, t in reversed(list(enumerate(tokens))):
            if t.spelling == "def" and tokens[i+1].spelling == "(":
                return (tokens[i+2].spelling == f"\"{func_name}\"")
        return False

    def get_insertion(self, is_init, tokens, is_staticmethod=False, cpp_func_name=None, param_count=None, tu_cursor=None, class_info=None):
        """Get the insertion string to fix a Python function signature.

        Args:
            is_init (bool): True if we are dealing with a constructor
            tokens (iterable): iterable of Token objects
            is_staticmethod (bool, optional): True if this a static method. Defaults to False.
            cpp_func_name (str, optional): C++ function name. Defaults to None.
            param_count (int, optional): expected number of parameters. Defaults to None.
            tu_cursor (Cursor, optional): translation unit cursor. Defaults to None.
            class_info (ClassInfo, optional): ClassInfo instance. Defaults to None.

        Returns:
            list(tuple)|None: list of tuples with source line number, source column number
            and string to be inserted on that line at that column position
        """
        insertion = self.find_python_args(tokens)
        log_path = self.cpp_path_noext + ".log"
        with open(log_path, "a") as hnd:
            print(f"1) get_insertion insertion {insertion}", file=hnd)
            hnd.flush()
        if insertion is not None:
            return insertion if not is_staticmethod else None
        insertion = self.find_python_arg(tokens)
        with open(log_path, "a") as hnd:
            print(f"2) get_insertion insertion {insertion}", file=hnd)
            hnd.flush()
        if insertion is not None:
            return insertion if not is_staticmethod else None
        insertion = self.find_no_arg(is_init, tokens, is_staticmethod, cpp_func_name, param_count, tu_cursor, class_info)
        with open(log_path, "a") as hnd:
            print(f"3) get_insertion insertion {insertion}", file=hnd)
            hnd.flush()
        return insertion

    def apply_insertions(self, insertions_by_line):
        """Apply insertions 

        Args:
            insertions_by_line (dict): dictionary of insertions keyed by line number (1-based)

        Returns:
            str: errors/warnings reported by clang-format
        """
        res = ""
        if not insertions_by_line:
            return res
        cpp_docmod_path = self.cpp_path_noext + FixSignatures.DOCMOD_SUFFIX + self.cpp_path_ext
        with open(self.cpp_path, "r") as cpp_in:
            with open(cpp_docmod_path, "w") as cpp_out:
                for line_num, line in enumerate(cpp_in):
                    insertions = insertions_by_line.get(line_num + 1, None)
                    if insertions is not None:
                        insertions_at_line = sorted(insertions, reverse=True)
                        for col_num, insertion in insertions_at_line:
                            col_num -= 1
                            line = line[:col_num] + insertion + line[col_num:]
                    cpp_out.write(line)
        with tempfile.NamedTemporaryFile() as clang_format_out:
            proc = subprocess.run([CLANG_FORMAT_EXE, cpp_docmod_path], stdout=clang_format_out.file, stderr=subprocess.PIPE)
            if proc.returncode:
                res += f"Failed to run {CLANG_FORMAT_EXE} on {cpp_docmod_path}"
            else:
                shutil.copyfile(clang_format_out.name, self.cpp_path)
                os.remove(cpp_docmod_path)
        if proc.stderr:
            msg = proc.stderr.decode("utf-8")
            res += f"{CLANG_FORMAT_EXE} reported the following on stderr:\n{msg}\n"
        return res

    @staticmethod
    def add_insertion(insertion, insertions_by_line):
        """Adds insertion to insertion dict.

        Args:
            insertion list(tuple): list of (line, col, text insertion) tuples
            insertions_by_line (dict): dictionary of insertions keyed by line number (1-based)
        """
        if insertion is not None:
            for ins in insertion:
                line, col, text = ins
                insertions_at_line = insertions_by_line.get(line, set())
                insertions_at_line.add((col, text))
                insertions_by_line[line] = insertions_at_line
        
    def parse_ast(self, arg1_func_byclass_dict):
        """Parse AST file generated by clang++.

        Args:
            arg1_func_byclass_dict (dict): dict relating class names to methods
            that need fixing. Also free functions are included under class name
            FixSignatures.NO_CLASS_KEY
        """
        self.assigned_overloads = {}
        try:
            translation_unit = TranslationUnit.from_ast_file(self.ast_path)
            out_path = self.cpp_path_noext + ".out"
            with open(out_path, "w") as hnd:
                self.recurse_ast_cursor(translation_unit.cursor, hnd)
            log_path = self.cpp_path_noext + ".log"
            with open(log_path, "w") as hnd:
                pass
            class_info_by_class_hash = self.find_nodes(translation_unit.cursor)
            class_method_node_hashes = set(itertools.chain.from_iterable([node.hash for node in class_info.parents] for class_info in class_info_by_class_hash.values()))
            arg1_non_class_func_names = arg1_func_byclass_dict.get(FixSignatures.NO_CLASS_KEY, None)
            if arg1_non_class_func_names is not None:
                non_class_defs = self.find_non_class_defs(translation_unit.cursor, class_method_node_hashes, arg1_non_class_func_names)
            with open(log_path, "a") as hnd:
                print(f"1) parse_ast cpp_path {self.cpp_path} class_info_by_class_hash {tuple(class_info_by_class_hash.keys())}", file=hnd)
                hnd.flush()
            class_info_by_class_name = self.prune_nodes(translation_unit.cursor, class_info_by_class_hash, arg1_func_byclass_dict)
            with open(log_path, "a") as hnd:
                print(f"2) parse_ast cpp_path {self.cpp_path} class_info_by_class_name {[(class_name, class_info.hash) for class_name, class_info in class_info_by_class_name.items()]}", file=hnd)
                hnd.flush()
            def_init_nodes_and_unassigned_func_names_by_class_name = self.find_def_init_nodes(
                class_info_by_class_name, arg1_func_byclass_dict)
            insertions = {}
            with open(log_path, "a") as hnd:
                print(f"3) parse_ast cpp_path {self.cpp_path} def_init_nodes_and_unassigned_func_names_by_class_name {def_init_nodes_and_unassigned_func_names_by_class_name}", file=hnd)
                hnd.flush()
            with open(log_path, "a") as hnd:
                for class_name, (def_init_nodes,_unassigned_func_names) in def_init_nodes_and_unassigned_func_names_by_class_name.items():
                    class_info = class_info_by_class_name[class_name]
                    for function_def in def_init_nodes:
                        tokens = list(function_def.def_cursor.get_tokens())
                        insertion = None
                        boost_python_entity = "".join(t.spelling for t in tokens[:3])
                        is_init = False
                        if boost_python_entity == "python::init":
                            print(f"4) parse_ast cpp_path {self.cpp_path} class_name {class_name} cpp_class_name {class_info.cpp_class_name} func_name {function_def.func_name} python::init tokens {[t.spelling for t in tokens]}", file=hnd)
                            hnd.flush()
                            is_init = True
                            insertion = self.get_insertion(is_init, tokens[3:], tu_cursor=translation_unit.cursor, class_info=class_info)
                        elif boost_python_entity == "python::class_":
                            res = self.find_cpp_func_r(function_def.def_cursor, function_def.level, function_def.func_name)
                            param_count = None
                            cpp_func_name = None
                            if res is not None:
                                param_count, cpp_func_name = res
                                print(f"5) parse_ast cpp_path {self.cpp_path} cpp_func_name {cpp_func_name} func_name {function_def.func_name} param_count {param_count} tokens {[t.spelling for t in tokens[3:]]}", file=hnd)
                                hnd.flush()
                            tokens_from_func_def = self.find_func_def(function_def.func_name, tokens[3:])
                            if tokens_from_func_def is not None:
                                print(f"6) parse_ast cpp_path {self.cpp_path} python::class_ tokens_from_func_def {[t.spelling for t in tokens_from_func_def]}", file=hnd)
                                hnd.flush()
                                insertion = self.get_insertion(is_init, tokens_from_func_def, function_def.is_staticmethod, cpp_func_name, param_count, translation_unit.cursor, class_info=class_info)
                        self.add_insertion(insertion, insertions)
                    print(f"8) parse_ast cpp_path {self.cpp_path} {insertions}", file=hnd)
                    hnd.flush()
                for func_name, def_nodes in non_class_defs.items():
                    for def_node in def_nodes:
                        tokens = list(def_node.get_tokens())
                        insertion = None
                        is_init = False
                        requested_level = 2
                        is_staticmethod = True
                        hnd.flush()
                        boost_python_entity = "".join(t.spelling for t in tokens[:3])
                        if boost_python_entity == "python::def":
                            res = self.find_cpp_func_r(def_node, requested_level, func_name)
                            param_count = None
                            cpp_func_name = None
                            if res is not None:
                                param_count, cpp_func_name = res
                                print(f"9) parse_ast cpp_path {self.cpp_path} cpp_func_name {cpp_func_name} func_name {func_name} param_count {param_count}", file=hnd)
                                hnd.flush()
                            tokens_from_func_def = self.find_func_def(func_name, tokens[2:])
                            if tokens_from_func_def is not None:
                                print(f"10) parse_ast cpp_path {self.cpp_path} python::def func_name {func_name} tokens_from_func_def {[t.spelling for t in tokens_from_func_def]}", file=hnd)
                                hnd.flush()
                                insertion = self.get_insertion(is_init, tokens_from_func_def, is_staticmethod, cpp_func_name, param_count, translation_unit.cursor)
                                self.add_insertion(insertion, insertions)
            self.ast_error += self.apply_insertions(insertions)
        except Exception as e:
            tb = traceback.format_exc()
            self.ast_error += f"{self.cpp_path}: Failed to parse AST\n{str(e)}\n{str(tb)}\n"


class ClangWorkerData(DictLike):
    """Data class passed to Worker as JSON string."""

    def __init__(self, clang_flags=None):
        self.clang_flags = clang_flags
        self.arg1_func_byclass_dict = {}


class FixSignatures:
    """Main FixSignatures class.

    Raises:
        ValueError
    """
    
    concurrency = max(1, multiprocessing.cpu_count() - 2)
    log_level = "INFO"
    cpp_source_path = os.environ.get("RDBASE", os.getcwd())
    rdkit_stubs_path = os.path.join(os.getcwd(), f"{RDKIT_MODULE_NAME}-stubs")
    clean = False
    include_path = os.path.join(os.environ.get("CONDA_PREFIX", os.getcwd()), "include")
    python_include_path = None
    rdkit_include_path = None
    clang_flags = "-emit-ast"
    user_clang_flags = ""
    CLANG_WORKER_SCRIPT = os.path.join(os.path.dirname(__file__), "clang_worker.py")
    DOCORIG_SUFFIX = "_RDKDOCORIG"
    DOCMOD_SUFFIX = "_RDKDOCMOD"
    DEFINE_RDK_REGEX = re.compile(r"^\s*#define\s+(RDK_\S+)\s*$")
    INCLUDE_PATH_BY_DEFINITION = {
        "RDK_BUILD_COORDGEN_SUPPORT": ["External", "CoordGen"],
        "RDK_USE_URF": ["External", "RingFamilies", "RingDecomposerLib", "src", "RingDecomposerLib"],
        "RDK_HAS_EIGEN3": os.environ.get("EIGEN3_INCLUDE_DIR", include_path),
        "RDK_BUILD_CAIRO_SUPPORT": [include_path, "cairo"],
    }
    NO_CLASS_KEY = "-"

    def __init__(self, args=None):
        """Constructor. Runs whole workflow.

        Args:
            args (Namespace, optional): ArgParser args
        """
        if args:
            for k, v in args._get_kwargs():
                setattr(self, k, v)
        self.logger = logging.getLogger(self.__class__.__name__)
        handler = logging.StreamHandler()
        formatter = logging.Formatter('[%(asctime)s %(levelname)s] %(message)s')
        handler.setFormatter(formatter)
        self.logger.addHandler(handler)
        self.logger.setLevel(getattr(logging, self.log_level))
        if self.python_include_path is None:
            python_include_path = sorted(glob.glob(os.path.join(self.include_path, "python*")))
            if python_include_path:
                self.python_include_path = python_include_path[0]
        if self.rdkit_include_path is None:
            print(f"Failed to find RDKit include path. Please set {self.__class__.__name__}.rdkit_include_path")
            sys.exit(1)
        self.init_cpp_file_dict()
        self.init_clang_worker_data()
        msg = self.generate_ast_files()
        if msg:
            self.logger.warning(msg)

    def init_cpp_file_dict(self):
        """Initialize dict relating each C++ file to patch to a CppFile object.
        Also creates backups of original C++ files with DOCORIG_SUFFIX extension
        if they do not exist yet. If they exist, it overwrites the current C++
        file with its backup. The --clean command line switch forces re-generating
        fresh backups from the current C++ file.

        Raises:
            ValueError
        """
        cpp_source_path = Path(self.cpp_source_path)
        self.cpp_file_dict = dict()
        paths = [p for p in sorted(cpp_source_path.rglob("*.cpp"))
                 if self.DOCMOD_SUFFIX not in str(p) and self.DOCORIG_SUFFIX not in str(p)
                 and "Demos" not in str(p)]
        cpp_paths_to_be_modified = []
        for p in paths:
            cpp_path = os.path.abspath(str(p))
            with open(cpp_path) as hnd:
                if any("python::class_" in line or "python::def" in line for line in hnd):
                    cpp_paths_to_be_modified.append(cpp_path)
                    cpp_path_noext, cpp_path_ext = os.path.splitext(cpp_path)
                    if cpp_path_noext in self.cpp_file_dict:
                        raise ValueError("There are multiple C++ files definining python::class_ "
                                         f"or python::def sharing the same basename {cpp_path_noext} "
                                         "but with different extensions; this should never happen")
                    self.cpp_file_dict[cpp_path_noext] = CppFile(cpp_path)
        for cpp_path in cpp_paths_to_be_modified:
            cpp_path_noext, cpp_path_ext = os.path.splitext(cpp_path)
            cpp_docorig_path = cpp_path_noext + self.DOCORIG_SUFFIX + cpp_path_ext
            have_docorig = os.path.exists(cpp_docorig_path)
            if self.clean and have_docorig:
                os.remove(cpp_docorig_path)
                have_docorig = False
            if not have_docorig:
                shutil.copyfile(cpp_path, cpp_docorig_path)
            else:
                shutil.copyfile(cpp_docorig_path, cpp_path)

    def get_rdk_build_flags(self):
        """Generate command line clang++ flags to build RDKit
        based on the contents of RDGeneral/RDConfig.h.

        Returns:
            str: command line clang++ flags
        """
        rdconfig_h = os.path.join(self.rdkit_include_path, RDKIT_MODULE_NAME, "RDGeneral", "RDConfig.h")
        definitions = set()
        includes = set()
        with open(rdconfig_h, "r") as hnd:
            for line in hnd:
                m = self.DEFINE_RDK_REGEX.match(line)
                if not m:
                    continue
                macro_name = m.group(1)
                definitions.add(macro_name)
                include_path = self.INCLUDE_PATH_BY_DEFINITION.get(macro_name, None)
                if include_path is None:
                    continue
                if not isinstance(include_path, str):
                    include_path = os.path.join(self.cpp_source_path, *include_path)
                includes.add(include_path)
        return (" ".join(f"-D{d}" for d in sorted(definitions)) +
                " " + " ".join(f"-I{i}" for i in sorted(includes)))

    @staticmethod
    def get_include_flags_from_include_path(include_path):
        """Generate command line clang++ include flags from include_path.

        Args:
            include_path (str): include path

        Returns:
            str: command line clang++ include flags
        """
        include_path = include_path or ""
        res = " ".join(f"-I{i}" for i in include_path.split(os.pathsep))
        if res:
            res = " " + res
        return res

    def add_func_to_dict_if_arg1(self, func, class_name=None):
        """Add the passed function to the dict of functions to be fixed if:
         1. it is a callable
         2. it has a docstring
         3. its docstring contains arg1

        Args:
            func (function): candidate function
            class_name (str, optional): clas name if the function is a class method
        """
        arg1_func_byclass_dict = self.clang_worker_data.arg1_func_byclass_dict
        if not isinstance(func.__doc__, str) or not callable(func) or "arg1" not in func.__doc__:
            return
        if class_name is None:
            class_name = self.NO_CLASS_KEY
        arg1_func_name_set = set(arg1_func_byclass_dict.get(class_name, []))
        arg1_func_name_set.add(func.__name__)
        arg1_func_byclass_dict[class_name] = sorted(arg1_func_name_set)

    def init_clang_worker_data(self):
        """Initialize ClangWorkerData."""
        rdkit_stubs_path = Path(self.rdkit_stubs_path)
        python_include_path = f"-I{self.python_include_path}" if self.python_include_path else ""
        rdkit_code = os.path.join(self.cpp_source_path, "Code")
        rdkit_external = os.path.join(self.cpp_source_path, "External")
        user_clang_flags = " " + self.user_clang_flags if self.user_clang_flags else ""
        rdk_build_defs = self.get_rdk_build_flags()
        qt_include_dirs = self.get_include_flags_from_include_path(os.environ.get("QT_INCLUDE_DIRS", None))
        rdkit_external_path = Path(rdkit_external)
        avalon_include_dir = os.path.abspath(str(max(rdkit_external_path.rglob("AvalonTools/ava-formake-AvalonToolkit_*/src/main/C/include"))))

        clang_flags = (
            f"-I{self.include_path} {python_include_path} -I{rdkit_code} "
            f"-I{rdkit_external} -I{avalon_include_dir} -I. -I..{qt_include_dirs}"
            f" {rdk_build_defs} {self.clang_flags}{user_clang_flags}"
        ).strip().split()
        self.clang_worker_data = ClangWorkerData(clang_flags)
        for p in sorted(rdkit_stubs_path.rglob("*.pyi")):
            if str(p.stem) == "__init__":
                p = p.parent
            pyi_module_path = os.path.splitext(str(p.relative_to(rdkit_stubs_path)).replace("/", "."))[0]
            if pyi_module_path == ".":
                pyi_module_path = RDKIT_MODULE_NAME
            else:
                pyi_module_path = RDKIT_MODULE_NAME + "." + pyi_module_path
            try:
                pyi_module = importlib.import_module(pyi_module_path)
            except Exception as e:
                self.logger.warning(f"ERROR: {str(e)}")
                continue
            for entry_name in dir(pyi_module):
                entry = getattr(pyi_module, entry_name, None)
                if (entry is None):
                    continue
                if entry.__class__.__name__ != "class":
                    self.add_func_to_dict_if_arg1(entry)
                else:
                    for method_name in dir(entry):
                        method = getattr(entry, method_name)
                        self.add_func_to_dict_if_arg1(method, entry_name)

    def clang_worker_thread(self, worker_idx):
        """Function run by each Worker thread.

        Args:
            worker_idx (int): Worker index (0-based)
        """
        proc = None
        res = WorkerResult(worker_idx)
        while 1:
            e = ""
            try:
                cpp_file_class = self.queue.get_nowait()
                self.logger.info(f"Processing {cpp_file_class.cpp_path}")
            except queue.Empty:
                self.logger.debug("Queue empty")
                break
            if proc is None:
                cmd = [sys.executable, self.CLANG_WORKER_SCRIPT, self.clang_worker_data.to_json()]
                try:
                    self.logger.debug(f"Attempting to run '{cmd}'")
                    proc = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                except Exception as e:
                    pass
            if proc is None:
                res.proc_error += f"Worker {worker_idx}: failed to start process.\n"
                if e:
                    res.proc_error += f"Exception was: {str(e)}\n"
                self.queue.task_done()
                continue
            stdout_data = None
            try:
                proc.stdin.write((cpp_file_class.to_json() + "\n").encode("utf-8"))
                proc.stdin.flush()
                stdout_data = proc.stdout.readline()
                if stdout_data:
                    stdout_data = stdout_data.decode("utf-8").strip()
                    cpp_file_class = CppFile.from_json(stdout_data)
                    self.cpp_file_dict[cpp_file_class.cpp_path_noext] = cpp_file_class
                    res.processed_cpp_files.add(cpp_file_class.cpp_path_noext)
                else:
                    proc.poll()
            except Exception:
                res.proc_error += (f"Exception while attempting to send {cpp_file_class.cpp_path} for processing "
                                   f"to {self.CLANG_WORKER_SCRIPT}:\n{str(e)}")
            self.queue.task_done()
            if not stdout_data and proc.returncode:
                res.proc_error += f"{self.CLANG_WORKER_SCRIPT} deamon not running."
                stderr_data = proc.stderr.read()
                if stderr_data:
                    res.proc_error += "\nError was:\n" + stderr_data.decode("utf-8")
                proc = None
                break
        if proc is not None:
            try:
                proc.stdin.write("\n".encode("utf-8"))
                proc.stdin.flush()
                stdout_data = proc.stdout.readline()
                if stdout_data:
                    stdout_data = stdout_data.decode("utf-8").strip()
                    if stdout_data:
                        res.proc_error += f"Worker {worker_idx}: expected empty message, found:\n{stdout_data}"
                else:
                    res.proc_error += f"Worker {worker_idx}: failed to receive empty message.\n"
            except Exception:
                pass
        self.thread_results[worker_idx] = res

    def generate_ast_files(self):
        """Generate clang++ AST files.

        Returns:
            str: errors generated by clang++
        """
        msg = ""
        self.queue = queue.Queue()
        cpp_class_files = list(self.cpp_file_dict.values())
        # Uncomment the following to troubleshoot specific file(s)
        # cpp_class_files = [f for f in cpp_class_files if os.path.basename(f.cpp_path) == "Atom.cpp"]
        n_files = len(cpp_class_files)
        self.logger.debug(f"Number of files: {n_files}")
        n_workers = min(self.concurrency, n_files)
        self.thread_results = {}
        clang_tasks = [Thread(target=self.clang_worker_thread, args=(i,), daemon=True) for i in range(n_workers)]
        for cpp_class_file in cpp_class_files:
            self.queue.put_nowait(cpp_class_file)
        for clang_task in clang_tasks:
            clang_task.start()
        have_alive_thread = True
        to_go_prev = n_files + 1
        while have_alive_thread:
            have_alive_thread = False
            for clang_task in clang_tasks:
                clang_task.join(timeout=0.1)
                have_alive_thread |= clang_task.is_alive()
            to_go_curr = [cpp_class.ast_error for cpp_class in cpp_class_files].count(None)
            if to_go_curr < to_go_prev:
                to_go_prev = to_go_curr
        for thread_idx, res in self.thread_results.items():
            if res.proc_error:
                msg += f"Process error in thread {thread_idx}:\n{res.proc_error}\n"
            for cpp_file_no_ext in sorted(res.processed_cpp_files):
                ast_error = self.cpp_file_dict[cpp_file_no_ext].ast_error
                if ast_error:
                    msg += f"clang AST errors in thread {thread_idx}:\n{ast_error}\n"
        return msg



# Copyright 2017, D. E. Shaw Research. All rights reserved.
#
#  This file is part of the RDKit.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the RDKit source tree.
#
import collections

from rdkit import Chem
from rdkit.Chem import AllChem


def GetHeterocycleReactionSmarts():
  """
    Return a list of the individual patterns for mutating individual
    atoms in aromatic rings to generate a new aromatic ring. The
    elements are collections.namedtuple objects with the following
    fields:

    SMARTS - the left side of the reaction SMARTS pattern, matches the atom to mutate
    CONVERT_FROM - the element type being converted: c, n, o, s
    CONVERT_TO - the right side of the reaction SMARTS pattern, there can be multiple destination types separated with a comma, these should map to multiple actual reaction objects
    EXAMPLE - an example aromatic ring system that SMARTS should match against, used in test cases
    NEGATIVE_EXAMPLE - an example aromatic ring system that SMART should NOT match against, used in test cases
    DESCRIPTION - a human readable description of the SMARTS pattern matching
    """
  HeteroAtomReaction = collections.namedtuple(
    'HeteroAtomReaction',
    ['SMARTS', 'CONVERT_FROM', 'CONVERT_TO', 'EXAMPLE', 'NEGATIVE_EXAMPLE', 'DESCRIPTION'])
  return [
    HeteroAtomReaction(
      "[c;x2;D2;r5:1]([n;x2;r5:2])[n;x2;r5:3]",
      "c",
      "[n:1]([n:2])[n:3]",
      "c1nc[nH]c1,c1ncn([*])c1,c1nncn([*])1",
      "",
      "aromatic carbon in 5 membered ring, bonded to 2 nitrogens",
    ),
    HeteroAtomReaction(
      "[c;x2;D2;r5:1]([n;h1;!+;!-;x2;r5;!$(*=*);!$(*[*;r5]=*):2])[n;x2;r5;!$(*=*);!$(*[*;r5]=*);!$(*[*;r5;x3]):3]",
      "c",
      "[o:1]([n;H0:2])[n;H0:3],[s:1]([n;H0:2])[n;H0:3]",
      "c1nc[nH]c1,[*]c1nc[nH]c1",
      "[*]c1ncno1,[*]n1[nH]cnc1=O,[*]c1nc2[nH]cnn2c(=O)c1[*],c1nc2cscc2[nH]1,c1c[nH]c2c1[nH]cn2,c1ccc2c(c1)[nH]cn2",
      "aromatic carbon in 5 membered ring, bonded to 1 nitrogen not bonded to another atom that is in 2 rings, bonded to 1 nitrogen with H and no charge",
    ),
    HeteroAtomReaction(
      "[c;x2;D2;r5:1]([n;h1;!+;!-;x2;r5;!$(*=*);!$(*[*;r5]=*):2])[n;x2;r5;!$(*=*);!$(*[*;r5]=*);!$(*[n;r5;x3]);$(*[c;x3;$(*[*;r6])]):3] ",
      "c",
      "[o:1]([n;H0:2])[n;H0:3],[s:1]([n;H0:2])[n;H0:3]",
      "c1ccc2c(c1)[nH]cn2",
      "[*]c1ncno1,[*]n1[nH]cnc1=O,[*]c1nc2[nH]cnn2c(=O)c1[*],c1nc2cscc2[nH]1,c1c[nH]c2c1[nH]cn2",
      "aromatic carbon in 5 membered ring, bonded to 1 nitrogen bonded to a bridgehead carbon in 6-member ring, bonded to 1 nitrogen with H and no charge",
    ),
    HeteroAtomReaction(
      "[c;x2;D2;r5:1]([c;x2;r5:2])[o,s;x2;r5:3]",
      "c",
      "[n:1]([c:2])[*:3]",
      "[*]c1ccoc1,c1ccoc1[*],n1ccoc1[*]",
      "",
      "aromatic carbon in 5 membered ring, bonded to 1 carbon and 1 oxygen or sulfur",
    ),
    HeteroAtomReaction(
      "[c;x2;D2;r5:1]([c;r5:2])[n;r5;h0;!-;!+:3]",
      "c",
      "[n:1]([c:2])[n:3]",
      "c1nc[nH]c1[*],n1cc[nH]c1[*],n1cc[nH][n+]1[*],[*]n1cc2c(c1)CCCC2,[*]N1Cc2cn([*])cc2C1,[*]c1c2c(cn1[*])CCCC2,[*]c1cn2c(c1[*])CCC2",
      "",
      "aromatic carbon in 5 membered ring, bonded to 1 carbon, bonded to 1 nitrogen with no H no charge",
    ),
    HeteroAtomReaction(
      "[c;x2;D2;r5:1]([c;x2;r5:2])[n;x2;r5;h1;+:3]",
      "c",
      "[n:1]([c:2])[n:3]",
      "c1c([*])sc[nH+]1,c1csc([*])[nH+]1",
      "",
      "aromatic carbon in 5 membered ring, bonded to 1 carbon, bonded to 1 nitrogen with H and charge",
    ),
    HeteroAtomReaction(
      "[c;x2;D2;r5:1]([c:2])[n;x2;r5;h1;!+;!-:3]",
      "c",
      "[n:1]([c:2])[n:3]",
      "[*]n1[nH]ccc1=O,[*]n1cc[nH]c1=O,[*]c1nc2[nH]ccn2c(=O)c1[*],[*]c1ccc[nH]1,[*]c1ncc[nH]1,[*][n+]1ccc[nH]1,[*][n+]1cc[nH]c1,c1ccc2c(c1)cc[nH]2",
      "[*]c1cn2ccnc2nc1[*]",
      "aromatic carbon in 5 membered ring, bonded to 1 carbon, bonded to 1 nitrogen with H and no charge",
    ),
    HeteroAtomReaction(
      "[c;x2;D2;r5:1]([c;x2;r5;!$(*=*);!$(*[*;r5]=*);!$(*[*;r5;x3]):2])[n;x2;r5;!$(*=*);!$(*[*;r5]=*);!$(*[*;r5;x3]);h1;!+;!-:3]",
      "c",
      "[o:1]([c:2])[n;H0:3],[s:1]([c:2])[n;H0:3]",
      "[*]c1ccc[nH]1,[*]c1ncc[nH]1,[*][n+]1ccc[nH]1,[*][n+]1cc[nH]c1",
      "c1ccc2c(c1)cc[nH]2,[*]n1[nH]ccc1=O,[*]n1cc[nH]c1=O,[*]c1nc2[nH]ccn2c(=O)c1[*],[*]c1cn2ccnc2nc1[*]",
      "aromatic carbon in 5 membered ring, bonded to 1 carbon, bonded to 1 nitrogen with H and no charge, no exocyclic double bonds or fused rings",
    ),
    HeteroAtomReaction(
      "[c;x2;D2;r5:1]([c;x2;r5:2])[n;h0;+;x2;r5:3]",
      "c",
      "[n:1]([c:2])[n:3]",
      "CC[n+]1ccco1,[*][n+]1ccco1,[*][n+]1ccc[nH]1",
      "",
      "aromatic carbon in 5 membered ring, bonded to 1 carbon, bonded to 1 nitrogen with no H and with charge",
    ),
    HeteroAtomReaction(
      "[c;x2;D2;r5:1]([c;x2;r5:2])[n;h0;+;x2;r5:3][n;x2;r5;h1:4]",
      "c",
      "[o:1]([c:2])[n:3][n;H0:4],[s:1]([c:2])[n:3][n;H0:4]",
      "[*][n+]1ccc[nH]1",
      "CC[n+]1ccco1,[*][n+]1ccco1",
      "aromatic carbon in 5 membered ring, bonded to 1 carbon, bonded to 1 nitrogen with no H and with charge, that nitrogen is then bonded to a nitrogen with a hydrogen",
    ),
    HeteroAtomReaction(
      "[c;x2;D2;r5:1]([c;x2;r5:2])[c;x2;r5:3][o,s;x2;r5:4]",
      "c",
      "[n:1]([c:2])[c:3][*:4]",
      "s1cccn1,o1c([*])ccn1",
      "s1cc([*])cn1",
      "aromatic carbon in 5 membered ring, bonded to 2 carbons, oxygen or sulfur in ring",
    ),
    HeteroAtomReaction(
      "[!o;!s;r5:1]1~[c;r5:2]~[c;x2;D2;r5:3]~[c;r5:4]~[!o;!s;r5:5]~1",
      "c",
      "[*:1]1~[c:2]~[n:3]~[c:4]~[*:5]~1",
      "[*]c1cc[nH]c1,[*]n1cccc1,[*]n1ccc2nccc-2c1",
      "",
      "aromatic carbon in 5 membered ring, bonded to 2 carbons, no oxygen or sulfur in ring",
    ),
    HeteroAtomReaction(
      "[n;h1;!+;!-;r5:1]1~[c;!$(*=*);r5:2]~[c;x2;D2;r5:3]~[c;!$(*=*);r5:4]~[!o;!s;r5:5]~1",
      "c",
      "[n;H0:1]1~[c:2]~[o:3]~[c:4]~[*:5]1,[n;H0:1]1~[c:2]~[s:3]~[c:4]~[*:5]~1",
      "c1cc[nH]c1,c1ccc2c(c1)cc[nH]2",
      "[*]n1cccc1,[*]n1cc([*])cc1,[*]n1ccc(=O)[nH]1",
      "aromatic carbon in 5 membered ring, bonded to 2 carbons without carbonyls, no oxygen or sulfur in ring, and a nitrogen with a hydrogen in the ring",
    ),
    HeteroAtomReaction(
      "[c;h1;D2;r6;$(a1ccc[c,n,o][c,n,o]1),$(a1cc[c,n,o][c,n,o]c1),$(a1c[c,n,o][c,n,o]cc1),$(a1[c,n,o][c,n,o]ccc1),$(a1cc[n,o]c[n,o]1),$(a1c[n,o]c[n,o]c1),$(a1[n,o]c[n,o]cc1),$(a1[n,o]ccc[n,o]1),$(a1c[n,o]cc[n,o]1),$(a1[n,o]cc[n,o]c1):1]",
      "c",
      "[n:1]",
      "c1ccccc1,c1cccnc1,c1nccnc1,[*]c1cc(=O)cco1,c1c([*])[nH]c(=O)[nH]c1=O,[*]c1c[nH]c(=O)[nH]c1=O",
      "",
      "aromatic carbon in 6 membered ring, has 1H, ring has less than 3 nitrogens",
    ),
    HeteroAtomReaction(
      "[n;x2;r5;h1;!+;!-;!$(*[o,s]):1]",
      "n",
      "[o:1],[s:1]",
      "c1n[nH]nn1,c1cc[nH]c1",
      "c1ccn([*])c1,[*]c1o[nH]c(=O)c1[*]",
      "aromatic nitrogen in 5 membered ring, 1H, no charge, not next to sulfur or oxygen",
    ),
    HeteroAtomReaction(
      "[n;x2;X2;r5;h0;!+;!-;$(*[a;r5][a;r5]):1]",
      "n",
      "[c:1]",
      "c1cnc[nH]1",
      "c1ccn([*])c1",
      "aromatic nitrogen in 5 membered ring, bonded to 2 heavy atoms, no H, no charge",
    ),
    HeteroAtomReaction(
      "",
      "n",
      "No mutations",
      "",
      "",
      "aromatic nitrogen in 5 membered ring, bonded to 3 heavy atoms, no charge",
    ),
    HeteroAtomReaction(
      "[n;x2;X3;r5;+;$(*[a;r5][a;r5]):1]",
      "n",
      "[c:1]",
      "c1coc[n+]1C",
      "",
      "aromatic nitrogen in 5 membered ring, bonded to 3 heavy atoms, charged",
    ),
    HeteroAtomReaction(
      "[n;x2;X2;r6;h0;!+;!-;$(*[a;r6][a;r6]):1]",
      "n",
      "[c:1]",
      "c1cccnc1,c1nccnc1",
      "",
      "aromatic nitrogen in 6 membered ring, bonded to 2 heavy atoms, no hydrogens, no charge",
    ),
    HeteroAtomReaction(
      "",
      "n",
      "No mutations",
      "",
      "",
      "aromatic nitrogen in 6 membered ring, bonded to 2 heavy atoms, 1H, no charge",
    ),
    HeteroAtomReaction(
      "[n;x2;r6;h1;+;$(*[a;r6][a;r6]):1]",
      "n",
      "[c:1]",
      "c1ccc2c(c1)ccc[nH+]2",
      "",
      "aromatic nitrogen in 6 membered ring, bonded to 2 heavy atoms, 1H, charged",
    ),
    HeteroAtomReaction(
      "[n;x2;r6;X3;+;$(*[a;r6][a;r6]):1]",
      "n",
      "[c:1]",
      "[*][n+]1cccc2c1cccc2",
      "",
      "aromatic nitrogen in 6 membered ring, bonded to 3 heavy atoms, charged",
    ),
    HeteroAtomReaction(
      "",
      "n",
      "No mutations",
      "",
      "",
      "aromatic nitrogen in 6 membered ring, bonded to 3 heavy atoms, no charge",
    ),
    HeteroAtomReaction(
      "[s;x2;r5;!+;!-;$(*([c;r5])[c;r5]):1]",
      "s",
      "[n;H1:1],[o:1]",
      "c1ccsc1,c1ncsc1",
      "",
      "aromatic sulfur in 5 membered ring, bonded to 2 carbons",
    ),
    HeteroAtomReaction(
      "[o;x2;r5;!+;!-;$(*([c;r5])[c;r5]):1]",
      "o",
      "[n;H1:1],[s:1]",
      "c1ccoc1,c1ncoc1",
      "",
      "aromatic oxygen in 5 membered ring, bonded to 2 carbons",
    ),
    HeteroAtomReaction(
      "[o,s;x2;r5;!+;!-:1]([c;x2;r5;!$(*=*):2])[n;x2;r5;h0;D2;!+;!-:3]",
      "o,s",
      "[c:1]([c:2])[n;H1:3]",
      "c1cnoc1,c1csnc1,c1c([*])snc1",
      "[*]n1occc1=O,[*][n+]1ccco1,[*]n1c(=O)[nH]sc1=O,[*]c1nsc(=O)o1",
      "aromatic oxygen or sulfur in 5 membered ring, bonded to 1 carbon, 1 nitrogen with no charge with no hydrogen and only bonded to 2 atoms",
    ),
    HeteroAtomReaction(
      "[o,s;x2;r5;!+;!-:1]([c;x2;r5:2])[n;x2;r5;+:3]",
      "o,s",
      "[c:1]([c:2])[n;+0:3]",
      "[*][n+]1ccco1",
      "",
      "aromatic oxygen or sulfur in 5 membered ring, bonded to 1 carbon, 1 nitrogen in ring with a charge",
    ),
    HeteroAtomReaction(
      "[s;x2;r5;!+;!-;$(*([c;r5])[n;h1;r5]):1]",
      "s",
      "[n;H1:1],[o:1]",
      "[*]c1s[nH]c(=O)c1[*]",
      "",
      "aromatic sulfur in 5 membered ring, bonded to 1 carbon, 1 nitrogen with a hydrogen",
    ),
    HeteroAtomReaction(
      "[o;x2;r5;!+;!-;$(*([c;r5])[n;h1;r5]):1]",
      "o",
      "[n;H1:1],[s:1]",
      "[*]c1o[nH]c(=O)c1[*]",
      "",
      "aromatic oxygen in 5 membered ring, bonded to 1 carbon, 1 nitrogen with a hydrogen",
    ),
    HeteroAtomReaction(
      "[s;x2;r5;!+;!-:1]([c;x2;r5:2])[n;x2;r5;h0;D3;!+;!-:3]",
      "s",
      "[n;H1:1]([c:2])[n:3],[o:1]([c:2])[n:3]",
      "[*]n1sc(=O)n([*])c1=O",
      "",
      "aromatic sulfur in 5 membered ring, bonded to 1 carbon, 1 nitrogen with no charge bonded to 3 atoms ",
    ),
    HeteroAtomReaction(
      "[o;x2;r5;!+;!-:1]([c;x2;r5:2])[n;x2;r5;h0;D3;!+;!-:3]",
      "o",
      "[n;H1:1]([c:2])[n:3],[s:1]([c:2])[n:3]",
      "[*]n1oc(=O)n([*])c1=O",
      "",
      "aromatic oxygen in 5 membered ring, bonded to 1 carbon, 1 nitrogen with no charge bonded to 3 atoms ",
    ),
    HeteroAtomReaction(
      "[o;r6:1]",
      "o",
      "[n;H1:1]",
      "[*]c1cccc(=O)o1",
      "",
      "aromatic oxygen in 6 membered ring",
    )
  ]


def _ParseReactions():
  for row in GetHeterocycleReactionSmarts():
    smarts = row.SMARTS
    if not smarts:
      continue

    for product in row.CONVERT_TO.split(','):
      reaction = smarts + '>>' + product
      yield AllChem.ReactionFromSmarts(reaction)


REACTION_CACHE = None


def GetHeterocycleReactions():
  """
    Return RDKit ChemicalReaction objects of the reaction SMARTS
    returned from GetHeterocyleReactionSmarts.
    """
  global REACTION_CACHE
  if REACTION_CACHE is None:
    REACTION_CACHE = list(_ParseReactions())
  return REACTION_CACHE


def EnumerateHeterocycles(inputmol, depth=None):
  """
    Enumerate possible relevant heterocycles on the given input
    molecule.

    >>> from rdkit.Chem.EnumerateHeterocycles import EnumerateHeterocycles
    >>> from rdkit import Chem
    >>> for smi in sorted(Chem.MolToSmiles(m) for m in EnumerateHeterocycles(Chem.MolFromSmiles('c1ccccc1'))):
    ...     print(smi)
    c1ccccc1
    c1ccncc1
    c1ccnnc1
    c1cnccn1
    c1cncnc1
    c1cnncn1
    c1cnnnc1
    c1ncncn1

    The algorithm works by mutating only one atom at a time. The depth
    parameter can be used to control the level of this recursion. For
    example, only enumerating aromatic rings that are one atom different:

    >>> for smi in sorted(Chem.MolToSmiles(m) for m in EnumerateHeterocycles(Chem.MolFromSmiles('n1ccccc1'), depth=1)):
    ...     print(smi)
    c1ccccc1
    c1ccnnc1
    c1cnccn1
    c1cncnc1
    """
  starting_points = [(inputmol, 0)]
  seen = set()
  while starting_points:
    curmol, curdepth = starting_points.pop(0)
    if depth is not None and curdepth >= depth:
      return

    for rxn in GetHeterocycleReactions():
      for newmol in rxn.RunReactants((curmol, )):
        newmol = newmol[0]
        Chem.SanitizeMol(newmol)
        newmol_smiles = Chem.MolToSmiles(newmol, isomericSmiles=True)
        if newmol_smiles in seen:
          continue

        starting_points.append((newmol, curdepth + 1))
        seen.add(newmol_smiles)
        yield newmol



import csv
import io
import logging
import os
import sys

try:
  import freewilson as fw
except:
  path = os.path.abspath(os.curdir)
  sys.path.insert(0, os.path.join(path, ".."))
  import freewilson as fw

import importlib
fw = importlib.reload(fw)

from rdkit import Chem, rdBase
from rdkit.Chem import rdFMCS, rdMolAlign

PATH = os.path.join(os.path.dirname(fw.__file__), 'data')
assert os.path.exists(PATH), PATH


def test_chembl():
  logging.getLogger().setLevel(logging.INFO)
  smilesfile = os.path.join(PATH, "CHEMBL2321810.smi")
  scaffoldfile = os.path.join(PATH, "CHEMBL2321810_scaffold.mol")
  csvfile = os.path.join(PATH, "CHEMBL2321810_act.csv")
  assert os.path.exists(smilesfile)
  mols = []
  for line in open(smilesfile):
    smiles, name = line.strip().split()
    m = Chem.MolFromSmiles(smiles)
    m.SetProp("_Name", name)
    mols.append(m)

  scaffold = Chem.MolFromMolBlock(open(scaffoldfile).read())
  data = {k: float(v) for k, v in list(csv.reader(open(csvfile)))[1:]}

  scores = [data[m.GetProp("_Name")] for m in mols]
  assert mols and len(mols) == len(scores)

  with rdBase.BlockLogs():
    free = fw.FWDecompose(scaffold, mols, scores)
  # let's make sure the r squared is decent
  assert free.r2 > 0.8, str(free.r2)

  # assert we get something
  preds = list(fw.FWBuild(free))
  assert len(preds)

  # check to see that the prediction filters work
  preds2 = list(fw.FWBuild(free, pred_filter=lambda x: x > 8))
  assert len(preds2)
  assert len([p for p in preds if p.prediction > 8]) == len(list(preds2))

  # check to see that the R groups are output in order, i.e. R10 after R3
  s = io.StringIO()
  fw.predictions_to_csv(s, free, preds2)
  assert s.getvalue()

  s2 = io.StringIO(s.getvalue())
  for i, row in enumerate(csv.reader(s2)):
    if i == 0:
      assert row == ['smiles', 'prediction', 'Core_smiles', 'R1_smiles', 'R3_smiles', 'R10_smiles']
  assert i > 0


def test_multicore():
  # test that we can add rgroups for later cores and not throw an exception
  scaffolds = [Chem.MolFromSmiles("c1ccccc1[*].NC=O"), Chem.MolFromSmiles("C1CCCCC1")]
  mols = [
    Chem.MolFromSmiles(x) for x in [
      'c1ccccc1CC2CNC2C(=O)N', 'Cc1ccccc1CC2CNC2C(=O)N', 'Cc1ccccc1CC2CNCC(=O)NC2',
      'C3c1ccccc1CC2CNC2C(=O)N3', 'C1CCCCC1F', 'ClC1CCCCC1F'
    ]
  ]
  decomp = fw.FWDecompose(scaffolds, mols, [1, 2, 3, 4, 5, 6])
  s = io.StringIO()
  fw.predictions_to_csv(s, decomp, fw.FWBuild(decomp))

def test_fep_benchmark_3D():
  benchmark = os.path.join(PATH, "cmet_ligands.sdf")
  mols = list(Chem.SDMolSupplier(benchmark, removeHs=False))
  smis = {Chem.MolToSmiles(m):m for m in mols if m}
  assert smis
  match = rdFMCS.FindMCS(mols)

  scores = [float(m.GetProp("r_exp_dg")) for m in mols]
  with rdBase.BlockLogs():
    free = fw.FWDecompose(match.queryMol, mols, scores)
  preds2 = list(fw.FWBuild(free, keep_training_set=True))
  assert preds2

  count = 0
  found = 0
  # we use an rms of 0.7 because we are assembling rgroups from different
  #  ligands together which may not exactly align
  for p in preds2:
    if not p.is_training: continue
    
    smi = Chem.MolToSmiles(p.mol)

    count += 1
    if smi not in smis: continue
    rms = rdMolAlign.GetBestRMS(smis[smi], p.mol)
    assert rms < 0.7





#!/usr/bin/env python

"""
patch_rdkit_docstrings is a runnable Python module.
It will scan RDKit sources searching for docstrings
that lack parameter definitions, or member functions that
do not have an explicit "self" parameter, and will patch
the C++ sources accordingly.

Example usage:

$ cd $RDBASE
$ CLANG_INCLUDE_PATH=/build/llvm-project/libcxx/include \
CLANG_PYTHON_BINDINGS_PATH=/build/llvm-project/clang/bindings/python \
QT_INCLUDE_DIRS=$CONDA_PREFIX/include/qt:$CONDA_PREFIX/include/qt/QtGui \
EIGEN3_INCLUDE_DIR=/usr/prog/Eigen/3.3.9-GCCcore-11.2.0/include \
python -m Scripts.patch_rdkit_docstrings
"""

import sys
import argparse
from . import FixSignatures
from ..gen_rdkit_stubs import purge_rdkit_source_dir_from_sys_path, find_rdkit_include_path, RDKIT_MODULE_NAME


def parse_args():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser()
    parser.add_argument("--cpp-source-path",
                        help="path where RDKit C++ sources are located (defaults to $RDBASE, or to cwd if RDBASE is not set)",
                        default=FixSignatures.cpp_source_path)
    parser.add_argument("--stubs-path",
                        help=f"path to the {RDKIT_MODULE_NAME}-stubs directory (defaults to ./{RDKIT_MODULE_NAME})",
                        default=FixSignatures.rdkit_stubs_path)
    parser.add_argument("--concurrency",
                        help=f"max number of CPUs to be used (defaults to {FixSignatures.concurrency})",
                        default=FixSignatures.concurrency)
    parser.add_argument("--include-path",
                        help=f"main clang include path (defaults to {FixSignatures.include_path})",
                        default=FixSignatures.include_path)
    parser.add_argument("--python-include-path",
                        help=f"Python clang include path (defaults to {FixSignatures.python_include_path})",
                        default=FixSignatures.python_include_path)
    parser.add_argument("--rdkit-include-path",
                        help=f"RDKit include path (defaults to {FixSignatures.rdkit_include_path})",
                        default=find_rdkit_include_path())
    parser.add_argument("--clang-flags",
                        help=f"flags to be passed to clang (defaults to {FixSignatures.clang_flags})",
                        default=FixSignatures.clang_flags)
    parser.add_argument("--user-clang-flags",
                        help=f"user-defined flags to be passed to clang (defaults to {FixSignatures.user_clang_flags})",
                        default=FixSignatures.clang_flags)
    parser.add_argument("--log-level",
                        help=f"logging level (defaults to {FixSignatures.log_level})",
                        default=FixSignatures.log_level)
    parser.add_argument("--clean", action="store_true", help="force removing all RDKDOCORIG files")
    args = parser.parse_args()
    return args

if __name__ == "__main__":
    args = parse_args()
    purge_rdkit_source_dir_from_sys_path()
    if args.rdkit_include_path is None:
        print(f"Failed to find RDKit include path. Please set it through the --rdkit-include-path switch.", file=sys.stderr)
        sys.exit(1)
    fix_signatures = FixSignatures(args)



#
# Copyright (C) 2002-2006 greg Landrum and Rational Discovery LLC
#
#   @@ All Rights Reserved @@
#  This file is part of the RDKit.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the RDKit source tree.
#
""" utility functionality for the 2D pharmacophores code

  See Docs/Chem/Pharm2D.triangles.jpg for an illustration of the way
  pharmacophores are broken into triangles and labelled.

  See Docs/Chem/Pharm2D.signatures.jpg for an illustration of bit
  numbering

"""
import itertools

try:
  # his is available in Python 3.8: https://docs.python.org/3/library/math.html
  from math import comb
except ImportError:
  from functools import lru_cache

  def _CheckCombArgument(n: int, k: int) -> None:
    if not isinstance(n, int) or not isinstance(k, int):
      raise ValueError(f"n ({n}) and k ({k}) must be positive integers")
    if n < 0:
      raise ValueError(f"n ({n}) must be a positive integer")
    if k < 0:
      raise ValueError(f"k ({k}) must be a positive integer")

  @lru_cache(maxsize=128)
  def comb(n: int, k: int) -> int:
    """
        Return the number of ways to choose k items from n items without repetition and without order.
        Evaluates to n! / (k! * (n - k)!) when k <= n and evaluates to zero when k > n.
        
        Arguments:
        ----------
            - n (integer): the number of items to choose from 
            - k (integer): the number of items to choose
        
        Returns:
        ----------
            - integer: the number of ways to choose k items from n items without repetition and without order
        
        
        Optimization reference:
        1) https://github.com/python/cpython/blob/main/Modules/mathmodule.c (Line 3381 <-> 3568)
        2) https://github.com/python/cpython/commit/60c320c38e4e95877cde0b1d8562ebd6bc02ac61
        3) https://bugs.python.org/issue37295 
        
        """
    _CheckCombArgument(n, k)
    if k > n:
      return 0
    if k == 0 or n == 0 or k == n:
      return 1
    if n - k > k:
      k = n - k

    # Optimization for small arguments
    if k < 8 or (16 * k < n < 512) or (k < n < 32):
      res = 1
      for i in range(k):
        res *= n - i
        res = res // (i + 1)
      return res

    j = k // 2
    return comb(n, j) * comb(n - j, k - j) // comb(k, j)


#
#  number of points in a scaffold -> sequence of distances (p1, p2) in
#   the scaffold
#
nPointDistDict = {
  2: ((0, 1), ),
  3: ((0, 1), (0, 2), (1, 2)),
  4: ((0, 1), (0, 2), (0, 3), (1, 2), (2, 3)),
  5: ((0, 1), (0, 2), (0, 3), (0, 4), (1, 2), (2, 3), (3, 4)),
  6: ((0, 1), (0, 2), (0, 3), (0, 4), (0, 5), (1, 2), (2, 3), (3, 4), (4, 5)),
  7: ((0, 1), (0, 2), (0, 3), (0, 4), (0, 5), (0, 6), (1, 2), (2, 3), (3, 4), (4, 5), (5, 6)),
  8: ((0, 1), (0, 2), (0, 3), (0, 4), (0, 5), (0, 6), (0, 7), (1, 2), (2, 3), (3, 4), (4, 5),
      (5, 6), (6, 7)),
  9: ((0, 1), (0, 2), (0, 3), (0, 4), (0, 5), (0, 6), (0, 7), (0, 8), (1, 2), (2, 3), (3, 4),
      (4, 5), (5, 6), (6, 7), (7, 8)),
  10: ((0, 1), (0, 2), (0, 3), (0, 4), (0, 5), (0, 6), (0, 7), (0, 8), (0, 9), (1, 2), (2, 3),
       (3, 4), (4, 5), (5, 6), (6, 7), (7, 8), (8, 9)),
}

#
#  number of distances in a scaffold -> number of points in the scaffold
#
nDistPointDict = {
  1: 2,
  3: 3,
  5: 4,
  7: 5,
  9: 6,
  11: 7,
  13: 8,
  15: 9,
  17: 10,
}

_trianglesInPharmacophore = {}


def GetTriangles(nPts):
  """ returns a tuple with the distance indices for
     triangles composing an nPts-pharmacophore

    """
  global _trianglesInPharmacophore
  if nPts < 3:
    return []
  res = _trianglesInPharmacophore.get(nPts, [])
  if not res:
    idx1, idx2, idx3 = (0, 1, nPts - 1)
    while idx1 < nPts - 2:
      res.append((idx1, idx2, idx3))
      idx1 += 1
      idx2 += 1
      idx3 += 1
    res = tuple(res)
    _trianglesInPharmacophore[nPts] = res
  return res


def BinsTriangleInequality(d1, d2, d3):
  """ checks the triangle inequality for combinations
      of distance bins.

      the general triangle inequality is:
         d1 + d2 >= d3
      the conservative binned form of this is:
         d1(upper) + d2(upper) >= d3(lower)

    """
  if d1[1] + d2[1] < d3[0]:
    return False
  if d2[1] + d3[1] < d1[0]:
    return False
  if d3[1] + d1[1] < d2[0]:
    return False
  return True


def ScaffoldPasses(combo, bins=None):
  """ checks the scaffold passed in to see if all
    contributing triangles can satisfy the triangle inequality

    the scaffold itself (encoded in combo) is a list of binned distances

    """
  # this is the number of points in the pharmacophore
  tris = GetTriangles(nDistPointDict[len(combo)])
  for tri in tris:
    ds = [bins[combo[x]] for x in tri]
    if not BinsTriangleInequality(ds[0], ds[1], ds[2]):
      return False
  return True


_numCombDict = {}


def NumCombinations(nItems, nSlots):
  """  returns the number of ways to fit nItems into nSlots

      We assume that (x, y) and (y, x) are equivalent, and
      (x, x) is allowed.

      General formula is, for N items and S slots:
        res = (N+S-1)! / ( (N-1)! * S! )

    """
  global _numCombDict
  res = _numCombDict.get((nItems, nSlots), -1)
  if res == -1:
    res = comb(nItems + nSlots - 1, nSlots)
    _numCombDict[(nItems, nSlots)] = res
  return res


_verbose = 0

_countCache = {}


def CountUpTo(nItems, nSlots, vs, idx=0, startAt=0):
  """ Figures out where a given combination of indices would
     occur in the combinatorial explosion generated by _GetIndexCombinations_

     **Arguments**

       - nItems: the number of items to distribute

       - nSlots: the number of slots in which to distribute them

       - vs: a sequence containing the values to find

       - idx: used in the recursion

       - startAt: used in the recursion

    **Returns**

       an integer

    """
  global _countCache
  if _verbose:
    print('  ' * idx, f'CountUpTo({idx})', vs[idx], startAt)
  if idx == 0 and (nItems, nSlots, tuple(vs)) in _countCache:
    return _countCache[(nItems, nSlots, tuple(vs))]
  elif idx >= nSlots:
    accum = 0
  elif idx == nSlots - 1:
    accum = vs[idx] - startAt
  else:
    accum = 0
    # get the digit at idx correct
    for i in range(startAt, vs[idx]):
      nLevsUnder = nSlots - idx - 1
      nValsOver = nItems - i
      numCombs = NumCombinations(nValsOver, nLevsUnder)
      if _verbose:
        print('  ' * idx, ' ', i, nValsOver, nLevsUnder, numCombs)
      accum += numCombs
    accum += CountUpTo(nItems, nSlots, vs, idx + 1, vs[idx])
  if _verbose:
    print('  ' * idx, '>', accum)
  if idx == 0:
    _countCache[(nItems, nSlots, tuple(vs))] = accum
  return accum


_indexCombinations = {}


def GetIndexCombinations(nItems, nSlots, slot=0, lastItemVal=0):
  """ Generates all combinations of nItems in nSlots without including
      duplicates

    **Arguments**

      - nItems: the number of items to distribute

      - nSlots: the number of slots in which to distribute them

      - slot: used in recursion

      - lastItemVal: used in recursion

    **Returns**

      a list of lists

    """
  global _indexCombinations
  if not slot and (nItems, nSlots) in _indexCombinations:
    res = _indexCombinations[(nItems, nSlots)]
  elif slot >= nSlots:
    res = []
  elif slot == nSlots - 1:
    res = [[x] for x in range(lastItemVal, nItems)]
  else:
    res = []
    for x in range(lastItemVal, nItems):
      tmp = GetIndexCombinations(nItems, nSlots, slot + 1, x)
      for entry in tmp:
        res.append([x] + entry)
    if not slot:
      _indexCombinations[(nItems, nSlots)] = res
  return res


def GetAllCombinations(choices, noDups=1, which=0):
  """  Does the combinatorial explosion of the possible combinations
    of the elements of _choices_.

    **Arguments**

      - choices: sequence of sequences with the elements to be enumerated

      - noDups: (optional) if this is nonzero, results with duplicates,
        e.g. (1,1,0), will not be generated

      - which: used in recursion

    **Returns**

      a list of lists

    >>> GetAllCombinations([(0, ), (1, ), (2, )])
    [[0, 1, 2]]
    >>> GetAllCombinations([(0, ), (1, 3), (2, )])
    [[0, 1, 2], [0, 3, 2]]

    >>> GetAllCombinations([(0, 1), (1, 3), (2, )])
    [[0, 1, 2], [0, 3, 2], [1, 3, 2]]

    """
  if which >= len(choices):
    return []
  elif which == len(choices) - 1:
    return [[x] for x in choices[which]]

  res = []
  tmp = GetAllCombinations(choices, noDups=noDups, which=which + 1)
  for thing in choices[which]:
    for other in tmp:
      if not noDups or thing not in other:
        res.append([thing] + other)
  return res


def GetUniqueCombinations(choices, classes):
  """  Does the combinatorial explosion of the possible combinations
    of the elements of _choices_.

    """
  assert len(choices) == len(classes)
  combos = set()
  for choice in itertools.product(*choices):
    # If a choice occurs in more than one of the fields, we ignore this case
    if len(set(choice)) != len(choice):
      continue
    combos.add(tuple(sorted((cls, ch) for cls, ch in zip(classes, choice))))
  return [list(combo) for combo in sorted(combos)]

GetUniqueCombinations_new = GetUniqueCombinations

def UniquifyCombinations(combos):
  """ uniquifies the combinations in the argument

      **Arguments**:

        - combos: a sequence of sequences

      **Returns**

        - a list of tuples containing the unique combos

    """
  resD = {}
  for combo in combos:
    k = combo[:]
    k.sort()
    resD[tuple(k)] = tuple(combo)
  return list(resD.values())


def GetPossibleScaffolds(nPts, bins, useTriangleInequality=True):
  """ gets all realizable scaffolds (passing the triangle inequality) with the
       given number of points and returns them as a list of tuples

    """
  if nPts < 2:
    return 0
  if nPts == 2:
    return [(x, ) for x in range(len(bins))]

  nDists = len(nPointDistDict[nPts])
  combos = GetAllCombinations([range(len(bins))] * nDists, noDups=0)
  return [
    tuple(combo) for combo in combos if not useTriangleInequality or ScaffoldPasses(combo, bins)
  ]


def OrderTriangle(featIndices, dists):
  """
      put the distances for a triangle into canonical order

      It's easy if the features are all different:

      >>> OrderTriangle([0, 2, 4], [1, 2, 3])
      ([0, 2, 4], [1, 2, 3])

      It's trickiest if they are all the same:
      
      >>> OrderTriangle([0, 0, 0], [1, 2, 3])
      ([0, 0, 0], [3, 2, 1])
      >>> OrderTriangle([0, 0, 0], [2, 1, 3])
      ([0, 0, 0], [3, 2, 1])
      >>> OrderTriangle([0, 0, 0], [1, 3, 2])
      ([0, 0, 0], [3, 2, 1])
      >>> OrderTriangle([0, 0, 0], [3, 1, 2])
      ([0, 0, 0], [3, 2, 1])
      >>> OrderTriangle([0, 0, 0], [3, 2, 1])
      ([0, 0, 0], [3, 2, 1])

      >>> OrderTriangle([0, 0, 1], [3, 2, 1])
      ([0, 0, 1], [3, 2, 1])
      >>> OrderTriangle([0, 0, 1], [1, 3, 2])
      ([0, 0, 1], [1, 3, 2])
      >>> OrderTriangle([0, 0, 1], [1, 2, 3])
      ([0, 0, 1], [1, 3, 2])

    """
  if len(featIndices) != 3:
    raise ValueError('bad indices')
  if len(dists) != 3:
    raise ValueError('bad dists')

  fs = set(featIndices)
  if len(fs) == 3:
    return featIndices, dists

  dSums = [0] * 3
  dSums[0] = dists[0] + dists[1]
  dSums[1] = dists[0] + dists[2]
  dSums[2] = dists[1] + dists[2]
  mD = max(dSums)
  if len(fs) == 1:
    if dSums[0] == mD:
      if dists[0] > dists[1]:
        ireorder = (0, 1, 2)
        dreorder = (0, 1, 2)
      else:
        ireorder = (0, 2, 1)
        dreorder = (1, 0, 2)
    elif dSums[1] == mD:
      if dists[0] > dists[2]:
        ireorder = (1, 0, 2)
        dreorder = (0, 2, 1)
      else:
        ireorder = (1, 2, 0)
        dreorder = (2, 0, 1)
    else:
      if dists[1] > dists[2]:
        ireorder = (2, 0, 1)
        dreorder = (1, 2, 0)
      else:
        ireorder = (2, 1, 0)
        dreorder = (2, 1, 0)
  else:
    # two classes
    if featIndices[0] == featIndices[1]:
      if dists[1] > dists[2]:
        ireorder = (0, 1, 2)
        dreorder = (0, 1, 2)
      else:
        ireorder = (1, 0, 2)
        dreorder = (0, 2, 1)
    elif featIndices[0] == featIndices[2]:
      if dists[0] > dists[2]:
        ireorder = (0, 1, 2)
        dreorder = (0, 1, 2)
      else:
        ireorder = (2, 1, 0)
        dreorder = (2, 1, 0)
    else:  # featIndices[1]==featIndices[2]:
      if dists[0] > dists[1]:
        ireorder = (0, 1, 2)
        dreorder = (0, 1, 2)
      else:
        ireorder = (0, 2, 1)
        dreorder = (1, 0, 2)

  dists = [dists[x] for x in dreorder]
  featIndices = [featIndices[x] for x in ireorder]
  return featIndices, dists


# ------------------------------------
#
#  doctest boilerplate
#
def _runDoctests(verbose=None):  # pragma: nocover
  import doctest
  import sys
  failed, _ = doctest.testmod(optionflags=doctest.ELLIPSIS, verbose=verbose)
  sys.exit(failed)


if __name__ == '__main__':  # pragma: nocover
  _runDoctests()



#
# Copyright (C) 2017 Peter Gedeck
#
#   @@ All Rights Reserved @@
#  This file is part of the RDKit.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the RDKit source tree.
#
'''
Collection of utilities to be used with descriptors

'''
import math


def setDescriptorVersion(version='1.0.0'):
  """ Set the version on the descriptor function.

  Use as a decorator """

  def wrapper(func):
    func.version = version
    return func

  return wrapper


class VectorDescriptorNamespace(dict):

  def __init__(self, **kwargs):
    self.update(kwargs)


class VectorDescriptorWrapper:
  """Wrap a function that returns a vector and make it seem like there
    is one function for each entry.  These functions are added to the global
    namespace with the names provided"""

  def __init__(self, func, names, version, namespace):
    self.func = func
    self.names = names
    self.func_key = "__%s" % (func.__name__)
    function_namespace = {}
    for i, n in enumerate(names):

      def f(mol, index=i):
        return self.call_desc(mol, index=index)

      f.__name__ = n
      f.__qualname__ = n
      f.version = version
      function_namespace[n] = f
    self.namespace = VectorDescriptorNamespace(**function_namespace)
    self.namespace.update(namespace)
    namespace.update(function_namespace)

  def _get_key(self, index):
    return "%s%s" % (self.func_key, index)

  def call_desc(self, mol, index):
    if hasattr(mol, self.func_key):
      results = getattr(mol, self.func_key, None)
      if results is not None:
        return results[index]

    try:
      results = self.func(mol)
    except Exception:
      return math.nan

    setattr(mol, self.func_key, results)
    return results[index]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# RDKit_unofficial_translation_JP documentation build configuration file, created by
# sphinx-quickstart on Sat Dec 29 12:41:54 2018.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'sphinxdoc'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
#html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
#html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
#html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = '../Images/logo.png'

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
#html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
#html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#html_use_smartypants = True

# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
html_sidebars = {'**': ['globaltoc.html', 'relations.html', 'searchbox.html']}

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
#html_domain_indices = True

# If false, no index is generated.
#html_use_index = True

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
#html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
#html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
#html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = 'RDKitdoc'

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['sphinx.ext.githubpages']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'RDKit_unofficial_translation_JP'
copyright = '2018, anonymous'
author = 'anonymous'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = '2019.03'
# The full version, including alpha/beta/rc tags.
release = '2019.03.1'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'ja'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
  # The paper size ('letterpaper' or 'a4paper').
  #
  # 'papersize': 'letterpaper',

  # The font size ('10pt', '11pt' or '12pt').
  #
  # 'pointsize': '10pt',

  # Additional stuff for the LaTeX preamble.
  #
  # 'preamble': '',

  # Latex figure (float) alignment
  #
  # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
  (master_doc, 'RDKit_unofficial_translation_JP.tex',
   'RDKit\\_unofficial\\_translation\\_JP Documentation', 'anonymous', 'manual'),
]

# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, 'rdkit_unofficial_translation_jp',
              'RDKit_unofficial_translation_JP Documentation', [author], 1)]

# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
  (master_doc, 'RDKit_unofficial_translation_JP', 'RDKit_unofficial_translation_JP Documentation',
   author, 'RDKit_unofficial_translation_JP', 'One line description of project.', 'Miscellaneous'),
]



#  Copyright (C) 2014-2024 Sereina Riniker and other RDKit contributors
#
#  This file is part of the RDKit.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the RDKit source tree.
#
""" Torsion Fingerprints (Deviation) (TFD)
    According to a paper from Schulz-Gasch et al., JCIM, 52, 1499-1512 (2012).

"""
import math
import os

from rdkit import Chem, Geometry, RDConfig, rdBase
from rdkit.Chem import rdchem, rdMolDescriptors
from rdkit.Chem import rdFingerprintGenerator

def _doMatch(inv, atoms):
  """ Helper function to check if all atoms in the list are the same
      
      Arguments:
      - inv:    atom invariants (used to define equivalence of atoms)
      - atoms:  list of atoms to check

      Return: boolean
  """
  match = True
  for i in range(len(atoms) - 1):
    for j in range(i + 1, len(atoms)):
      if (inv[atoms[i].GetIdx()] != inv[atoms[j].GetIdx()]):
        match = False
        return match
  return match


def _doNotMatch(inv, atoms):
  """ Helper function to check if all atoms in the list are NOT the same
      
      Arguments:
      - inv:    atom invariants (used to define equivalence of atoms)
      - atoms:  list of atoms to check

      Return: boolean
  """
  match = True
  for i in range(len(atoms) - 1):
    for j in range(i + 1, len(atoms)):
      if (inv[atoms[i].GetIdx()] == inv[atoms[j].GetIdx()]):
        match = False
        return match
  return match


def _doMatchExcept1(inv, atoms):
  """ Helper function to check if two atoms in the list are the same, 
      and one not
      Note: Works only for three atoms
      
      Arguments:
      - inv:    atom invariants (used to define equivalence of atoms)
      - atoms:  list of atoms to check

      Return: atom that is different
  """
  if len(atoms) != 3:
    raise ValueError("Number of atoms must be three")
  a1 = atoms[0].GetIdx()
  a2 = atoms[1].GetIdx()
  a3 = atoms[2].GetIdx()
  if (inv[a1] == inv[a2] and inv[a1] != inv[a3] and inv[a2] != inv[a3]):
    return atoms[2]
  elif (inv[a1] != inv[a2] and inv[a1] == inv[a3] and inv[a2] != inv[a3]):
    return atoms[1]
  elif (inv[a1] != inv[a2] and inv[a1] != inv[a3] and inv[a2] == inv[a3]):
    return atoms[0]
  return None


def _getAtomInvariantsWithRadius(mol, radius):
  """ Helper function to calculate the atom invariants for each atom 
      with a given radius

      Arguments:
      - mol:    the molecule of interest
      - radius: the radius for the Morgan fingerprint

      Return: list of atom invariants
  """
  inv = []
  fpg = rdFingerprintGenerator.GetMorganGenerator(radius,includeRedundantEnvironments=True)
  ao = rdFingerprintGenerator.AdditionalOutput()
  ao.AllocateBitInfoMap()

  fp = fpg.GetSparseCountFingerprint(mol, additionalOutput=ao)
  bim = ao.GetBitInfoMap()
  inv = [0]*mol.GetNumAtoms()
  for k,vs in bim.items():
    for (aid,r) in vs:
      if r == radius:
        inv[aid] = k
  return inv


def _getHeavyAtomNeighbors(atom1, aid2=-1):
  """ Helper function to calculate the number of heavy atom neighbors.

      Arguments:
      - atom1:    the atom of interest
      - aid2:     atom index that should be excluded from neighbors (default: none)

      Return: a list of heavy atom neighbors of the given atom
  """
  if aid2 < 0:
    return [n for n in atom1.GetNeighbors() if n.GetSymbol() != 'H']
  else:
    return [n for n in atom1.GetNeighbors() if (n.GetSymbol() != 'H' and n.GetIdx() != aid2)]


def _getIndexforTorsion(neighbors, inv):
  """ Helper function to calculate the index of the reference atom for 
      a given atom

      Arguments:
      - neighbors:  list of the neighbors of the atom
      - inv:        atom invariants

      Return: list of atom indices as reference for torsion
  """
  if len(neighbors) == 1:  # atom has only one neighbor
    return [neighbors[0]]
  elif _doMatch(inv, neighbors):  # atom has all symmetric neighbors
    return neighbors
  elif _doNotMatch(inv, neighbors):  # atom has all different neighbors
    # sort by atom inv and simply use the first neighbor
    neighbors = sorted(neighbors, key=lambda x: inv[x.GetIdx()])
    return [neighbors[0]]
  elif len(neighbors) == 3:
    at = _doMatchExcept1(inv, neighbors)  # two neighbors the same, one different
    if at is None:
      raise ValueError("Atom neighbors are either all the same or all different")
    return [at]
  else:  # weird case
    # sort by atom inv and simply use the first neighbor
    neighbors = sorted(neighbors, key=lambda x: inv[x.GetIdx()])
    return [neighbors[0]]


def _getBondsForTorsions(mol, ignoreColinearBonds):
  """ Determine the bonds (or pair of atoms treated like a bond) for which
      torsions should be calculated.

      Arguments:
      - refmol: the molecule of interest
      - ignoreColinearBonds: if True (default), single bonds adjacent to
                             triple bonds are ignored
                             if False, alternative not-covalently bound
                             atoms are used to define the torsion
  """
  # flag the atoms that cannot be part of the centre atoms of a torsion
  # patterns: triple bonds and allenes
  patts = [Chem.MolFromSmarts(x) for x in ['*#*', '[$([C](=*)=*)]']]
  atomFlags = [0] * mol.GetNumAtoms()
  for p in patts:
    if mol.HasSubstructMatch(p):
      matches = mol.GetSubstructMatches(p)
      for match in matches:
        for a in match:
          atomFlags[a] = 1

  bonds = []
  doneBonds = [0] * mol.GetNumBonds()
  for b in mol.GetBonds():
    if b.IsInRing():
      continue
    a1 = b.GetBeginAtomIdx()
    a2 = b.GetEndAtomIdx()
    nb1 = _getHeavyAtomNeighbors(b.GetBeginAtom(), a2)
    nb2 = _getHeavyAtomNeighbors(b.GetEndAtom(), a1)
    if not doneBonds[b.GetIdx()] and (nb1 and nb2):  # no terminal bonds
      doneBonds[b.GetIdx()] = 1
      # check if atoms cannot be middle atoms
      if atomFlags[a1] or atomFlags[a2]:
        if not ignoreColinearBonds:  # search for alternative not-covalently bound atoms
          while len(nb1) == 1 and atomFlags[a1]:
            a1old = a1
            a1 = nb1[0].GetIdx()
            b = mol.GetBondBetweenAtoms(a1old, a1)
            if b.GetEndAtom().GetIdx() == a1old:
              nb1 = _getHeavyAtomNeighbors(b.GetBeginAtom(), a1old)
            else:
              nb1 = _getHeavyAtomNeighbors(b.GetEndAtom(), a1old)
            doneBonds[b.GetIdx()] = 1
          while len(nb2) == 1 and atomFlags[a2]:
            doneBonds[b.GetIdx()] = 1
            a2old = a2
            a2 = nb2[0].GetIdx()
            b = mol.GetBondBetweenAtoms(a2old, a2)
            if b.GetBeginAtom().GetIdx() == a2old:
              nb2 = _getHeavyAtomNeighbors(b.GetEndAtom(), a2old)
            else:
              nb2 = _getHeavyAtomNeighbors(b.GetBeginAtom(), a2old)
            doneBonds[b.GetIdx()] = 1
          if nb1 and nb2:
            bonds.append((a1, a2, nb1, nb2))

      else:
        bonds.append((a1, a2, nb1, nb2))
  return bonds


def CalculateTorsionLists(mol, maxDev='equal', symmRadius=2, ignoreColinearBonds=True):
  """ Calculate a list of torsions for a given molecule. For each torsion
      the four atom indices are determined and stored in a set.

      Arguments:
      - mol:      the molecule of interest
      - maxDev:   maximal deviation used for normalization
                  'equal': all torsions are normalized using 180.0 (default)
                  'spec':  each torsion is normalized using its specific
                           maximal deviation as given in the paper
      - symmRadius: radius used for calculating the atom invariants
                    (default: 2)
      - ignoreColinearBonds: if True (default), single bonds adjacent to
                             triple bonds are ignored
                             if False, alternative not-covalently bound
                             atoms are used to define the torsion

      Return: two lists of torsions: non-ring and ring torsions
  """
  if maxDev not in ['equal', 'spec']:
    raise ValueError("maxDev must be either equal or spec")
  # get non-terminal, non-cyclic bonds
  bonds = _getBondsForTorsions(mol, ignoreColinearBonds)
  # get atom invariants
  if symmRadius > 0:
    inv = _getAtomInvariantsWithRadius(mol, symmRadius)
  else:
    inv = rdMolDescriptors.GetConnectivityInvariants(mol)
  # get the torsions
  tors_list = []  # to store the atom indices of the torsions
  for a1, a2, nb1, nb2 in bonds:
    d1 = _getIndexforTorsion(nb1, inv)
    d2 = _getIndexforTorsion(nb2, inv)
    if len(d1) == 1 and len(d2) == 1:  # case 1, 2, 4, 5, 7, 10, 16, 12, 17, 19
      tors_list.append(([(d1[0].GetIdx(), a1, a2, d2[0].GetIdx())], 180.0))
    elif len(d1) == 1:  # case 3, 6, 8, 13, 20
      if len(nb2) == 2:  # two neighbors
        tors_list.append(([(d1[0].GetIdx(), a1, a2, nb.GetIdx()) for nb in d2], 90.0))
      else:  # three neighbors
        tors_list.append(([(d1[0].GetIdx(), a1, a2, nb.GetIdx()) for nb in d2], 60.0))
    elif len(d2) == 1:  # case 3, 6, 8, 13, 20
      if len(nb1) == 2:
        tors_list.append(([(nb.GetIdx(), a1, a2, d2[0].GetIdx()) for nb in d1], 90.0))
      else:  # three neighbors
        tors_list.append(([(nb.GetIdx(), a1, a2, d2[0].GetIdx()) for nb in d1], 60.0))
    else:  # both symmetric
      tmp = []
      for n1 in d1:
        for n2 in d2:
          tmp.append((n1.GetIdx(), a1, a2, n2.GetIdx()))
      if len(nb1) == 2 and len(nb2) == 2:  # case 9
        tors_list.append((tmp, 90.0))
      elif len(nb1) == 3 and len(nb2) == 3:  # case 21
        tors_list.append((tmp, 60.0))
      else:  # case 15
        tors_list.append((tmp, 30.0))
  # maximal possible deviation for non-cyclic bonds
  if maxDev == 'equal':
    tors_list = [(t, 180.0) for t, d in tors_list]
  # rings
  rings = Chem.GetSymmSSSR(mol)
  tors_list_rings = []
  for r in rings:
    # get the torsions
    tmp = []
    num = len(r)
    if 14 <= num:
      maxdev = 180.0
    else:
      maxdev = 180.0 * math.exp(-0.025 * (num - 14) * (num - 14))
    for i in range(len(r)):
      tmp.append((r[i], r[(i + 1) % num], r[(i + 2) % num], r[(i + 3) % num]))
    tors_list_rings.append((tmp, maxdev))
  return tors_list, tors_list_rings


def _getTorsionAtomPositions(atoms, conf):
  """ Helper function to retrieve the coordinates of the four atoms
      in a torsion

      Arguments:
      - atoms:   list with the four atoms
      - conf:    conformation of the molecule

      Return: Point3D objects of the four atoms
  """
  if len(atoms) != 4:
    raise ValueError("List must contain exactly four atoms")
  p1 = conf.GetAtomPosition(atoms[0])
  p2 = conf.GetAtomPosition(atoms[1])
  p3 = conf.GetAtomPosition(atoms[2])
  p4 = conf.GetAtomPosition(atoms[3])
  return p1, p2, p3, p4


def CalculateTorsionAngles(mol, tors_list, tors_list_rings, confId=-1):
  """ Calculate the torsion angles for a list of non-ring and 
      a list of ring torsions.

      Arguments:
      - mol:       the molecule of interest
      - tors_list: list of non-ring torsions
      - tors_list_rings: list of ring torsions
      - confId:    index of the conformation (default: first conformer)

      Return: list of torsion angles
  """
  torsions = []
  conf = mol.GetConformer(confId)
  for quartets, maxdev in tors_list:
    tors = []
    # loop over torsions and calculate angle
    for atoms in quartets:
      p1, p2, p3, p4 = _getTorsionAtomPositions(atoms, conf)
      tmpTors = (Geometry.ComputeSignedDihedralAngle(p1, p2, p3, p4) / math.pi) * 180.0
      if tmpTors < 0:
        tmpTors += 360.0  # angle between 0 and 360
      tors.append(tmpTors)
    torsions.append((tors, maxdev))
  # rings
  for quartets, maxdev in tors_list_rings:
    num = len(quartets)
    # loop over torsions and sum them up
    tors = 0
    for atoms in quartets:
      p1, p2, p3, p4 = _getTorsionAtomPositions(atoms, conf)
      tmpTors = abs((Geometry.ComputeSignedDihedralAngle(p1, p2, p3, p4) / math.pi) * 180.0)
      tors += tmpTors
    tors /= num
    torsions.append(([tors], maxdev))
  return torsions


def _findCentralBond(mol, distmat):
  """ Helper function to identify the atoms of the most central bond.

      Arguments:
      - mol:     the molecule of interest
      - distmat: distance matrix of the molecule

      Return: atom indices of the two most central atoms (in order)
  """
  from numpy import std

  # get the most central atom = atom with the least STD of shortest distances
  stds = []
  for i in range(mol.GetNumAtoms()):
    # only consider non-terminal atoms
    if len(_getHeavyAtomNeighbors(mol.GetAtomWithIdx(i))) < 2:
      continue
    tmp = [d for d in distmat[i]]
    tmp.pop(i)
    stds.append((std(tmp), i))
  stds.sort()
  aid1 = stds[0][1]
  # find the second most central bond that is bonded to aid1
  i = 1
  while 1:
    if mol.GetBondBetweenAtoms(aid1, stds[i][1]) is None:
      i += 1
    else:
      aid2 = stds[i][1]
      break
  return aid1, aid2  # most central atom comes first


def _calculateBeta(mol, distmat, aid1):
  """ Helper function to calculate the beta for torsion weights
      according to the formula in the paper.
      w(dmax/2) = 0.1

      Arguments:
      - mol:     the molecule of interest
      - distmat: distance matrix of the molecule
      - aid1:    atom index of the most central atom

      Return: value of beta (float)
  """
  # get all non-terminal bonds
  bonds = []
  for b in mol.GetBonds():
    nb1 = _getHeavyAtomNeighbors(b.GetBeginAtom())
    nb2 = _getHeavyAtomNeighbors(b.GetEndAtom())
    if len(nb1) > 1 and len(nb2) > 1:
      bonds.append(b)
  # get shortest distance
  dmax = 0
  for b in bonds:
    bid1 = b.GetBeginAtom().GetIdx()
    bid2 = b.GetEndAtom().GetIdx()
    d = max([distmat[aid1][bid1], distmat[aid1][bid2]])
    if (d > dmax):
      dmax = d
  dmax2 = dmax / 2.0
  beta = -math.log(0.1) / (dmax2 * dmax2)
  return beta


def CalculateTorsionWeights(mol, aid1=-1, aid2=-1, ignoreColinearBonds=True):
  """ Calculate the weights for the torsions in a molecule.
      By default, the highest weight is given to the bond 
      connecting the two most central atoms.
      If desired, two alternate atoms can be specified (must 
      be connected by a bond).

      Arguments:
      - mol:   the molecule of interest
      - aid1:  index of the first atom (default: most central)
      - aid2:  index of the second atom (default: second most central)
      - ignoreColinearBonds: if True (default), single bonds adjacent to
                             triple bonds are ignored
                             if False, alternative not-covalently bound
                             atoms are used to define the torsion

      Return: list of torsion weights (both non-ring and ring)
  """
  # get distance matrix
  distmat = Chem.GetDistanceMatrix(mol)
  if aid1 < 0 and aid2 < 0:
    aid1, aid2 = _findCentralBond(mol, distmat)
  else:
    b = mol.GetBondBetweenAtoms(aid1, aid2)
    if b is None:
      raise ValueError("Specified atoms must be connected by a bond.")
  # calculate beta according to the formula in the paper
  beta = _calculateBeta(mol, distmat, aid1)
  # get non-terminal, non-cyclic bonds
  bonds = _getBondsForTorsions(mol, ignoreColinearBonds)
  # get shortest paths and calculate weights
  weights = []
  for bid1, bid2, nb1, nb2 in bonds:
    if ((bid1, bid2) == (aid1, aid2)
        or (bid2, bid1) == (aid1, aid2)):  # if it's the most central bond itself
      d = 0
    else:
      # get shortest distance between the 4 atoms and add 1 to get bond distance
      d = min(distmat[aid1][bid1], distmat[aid1][bid2], distmat[aid2][bid1],
              distmat[aid2][bid2]) + 1
    w = math.exp(-beta * (d * d))
    weights.append(w)

  ## RINGS
  rings = mol.GetRingInfo()
  for r in rings.BondRings():
    # get shortest distances
    tmp = []
    num = len(r)
    for bidx in r:
      b = mol.GetBondWithIdx(bidx)
      bid1 = b.GetBeginAtomIdx()
      bid2 = b.GetEndAtomIdx()
      # get shortest distance between the 4 atoms and add 1 to get bond distance
      d = min(distmat[aid1][bid1], distmat[aid1][bid2], distmat[aid2][bid1],
              distmat[aid2][bid2]) + 1
      tmp.append(d)
    # calculate weights and append to list
    # Note: the description in the paper is not very clear, the following
    #       formula was found to give the same weights as shown in Fig. 1
    #       For a ring of size N: w = N/2 * exp(-beta*(sum(w of each bond in ring)/N)^2)
    w = sum(tmp) / float(num)
    w = math.exp(-beta * (w * w))
    weights.append(w * (num / 2.0))
  return weights


def CalculateTFD(torsions1, torsions2, weights=None):
  """ Calculate the torsion deviation fingerprint (TFD) given two lists of
      torsion angles.

      Arguments:
      - torsions1:  torsion angles of conformation 1
      - torsions2:  torsion angles of conformation 2
      - weights:    list of torsion weights (default: None)

      Return: TFD value (float)
  """
  if len(torsions1) != len(torsions2):
    raise ValueError("List of torsions angles must have the same size.")
  # calculate deviations and normalize (divide by max. possible deviation)
  deviations = []
  for tors1, tors2 in zip(torsions1, torsions2):
    mindiff = 180.0
    for t1 in tors1[0]:
      for t2 in tors2[0]:
        diff = abs(t1 - t2)
        if (360.0 - diff) < diff:  # we do not care about direction
          diff = 360.0 - diff
        #print t1, t2, diff
        if diff < mindiff:
          mindiff = diff
    deviations.append(mindiff / tors1[1])
  # do we use weights?
  if weights is not None:
    if len(weights) != len(torsions1):
      raise ValueError("List of torsions angles and weights must have the same size.")
    deviations = [d * w for d, w in zip(deviations, weights)]
    sum_weights = sum(weights)
  else:
    sum_weights = len(deviations)
  tfd = sum(deviations)
  if sum_weights != 0:  # avoid division by zero
    tfd /= sum_weights
  return tfd


def _getSameAtomOrder(mol1, mol2):
  """ Generate a new molecule with the atom order of mol1 and coordinates
      from mol2.
      
      Arguments:
      - mol1:     first instance of the molecule of interest
      - mol2:     second instance the molecule of interest

      Return: RDKit molecule
  """
  match = mol2.GetSubstructMatch(mol1)
  atomNums = tuple(range(mol1.GetNumAtoms()))
  if match != atomNums:  # atom orders are not the same!
    #print "Atoms of second molecule reordered."
    mol3 = Chem.Mol(mol1)
    mol3.RemoveAllConformers()
    for conf2 in mol2.GetConformers():
      confId = conf2.GetId()
      conf = rdchem.Conformer(mol1.GetNumAtoms())
      conf.SetId(confId)
      for i in range(mol1.GetNumAtoms()):
        conf.SetAtomPosition(i, mol2.GetConformer(confId).GetAtomPosition(match[i]))
      cid = mol3.AddConformer(conf)
    return mol3
  else:
    return Chem.Mol(mol2)


# some wrapper functions
def GetTFDBetweenConformers(mol, confIds1, confIds2, useWeights=True, maxDev='equal', symmRadius=2,
                            ignoreColinearBonds=True):
  """ Wrapper to calculate the TFD between two list of conformers 
      of a molecule

      Arguments:
      - mol:      the molecule of interest
      - confIds1:  first list of conformer indices
      - confIds2:  second list of conformer indices
      - useWeights: flag for using torsion weights in the TFD calculation
      - maxDev:   maximal deviation used for normalization
                  'equal': all torsions are normalized using 180.0 (default)
                  'spec':  each torsion is normalized using its specific
                           maximal deviation as given in the paper
      - symmRadius: radius used for calculating the atom invariants
                    (default: 2)
      - ignoreColinearBonds: if True (default), single bonds adjacent to
                             triple bonds are ignored
                             if False, alternative not-covalently bound
                             atoms are used to define the torsion

      Return: list of TFD values
  """
  tl, tlr = CalculateTorsionLists(mol, maxDev=maxDev, symmRadius=symmRadius,
                                  ignoreColinearBonds=ignoreColinearBonds)
  torsions1 = [CalculateTorsionAngles(mol, tl, tlr, confId=cid) for cid in confIds1]
  torsions2 = [CalculateTorsionAngles(mol, tl, tlr, confId=cid) for cid in confIds2]
  tfd = []
  if useWeights:
    weights = CalculateTorsionWeights(mol, ignoreColinearBonds=ignoreColinearBonds)
  else:
    weights = None
  for t1 in torsions1:
    for t2 in torsions2:
      tfd.append(CalculateTFD(t1, t2, weights=weights))
  return tfd


def GetTFDBetweenMolecules(mol1, mol2, confId1=-1, confId2=-1, useWeights=True, maxDev='equal',
                           symmRadius=2, ignoreColinearBonds=True):
  """ Wrapper to calculate the TFD between two molecules.
      Important: The two molecules must be isomorphic

      Arguments:
      - mol1:     first instance of the molecule of interest
      - mol2:     second instance the molecule of interest
      - confId1:  conformer index for mol1 (default: first conformer)
      - confId2:  conformer index for mol2 (default: first conformer)
      - useWeights: flag for using torsion weights in the TFD calculation
      - maxDev:   maximal deviation used for normalization
                  'equal': all torsions are normalized using 180.0 (default)
                  'spec':  each torsion is normalized using its specific
                           maximal deviation as given in the paper
      - symmRadius: radius used for calculating the atom invariants
                    (default: 2)
      - ignoreColinearBonds: if True (default), single bonds adjacent to
                             triple bonds are ignored
                             if False, alternative not-covalently bound
                             atoms are used to define the torsion

      Return: TFD value
  """
  if (Chem.MolToSmiles(mol1) != Chem.MolToSmiles(mol2)):
    raise ValueError("The two molecules must be instances of the same molecule!")
  mol2 = _getSameAtomOrder(mol1, mol2)
  tl, tlr = CalculateTorsionLists(mol1, maxDev=maxDev, symmRadius=symmRadius,
                                  ignoreColinearBonds=ignoreColinearBonds)
  # first molecule
  torsion1 = CalculateTorsionAngles(mol1, tl, tlr, confId=confId1)
  # second molecule
  torsion2 = CalculateTorsionAngles(mol2, tl, tlr, confId=confId2)
  if useWeights:
    weights = CalculateTorsionWeights(mol1, ignoreColinearBonds=ignoreColinearBonds)
  else:
    weights = None
  tfd = CalculateTFD(torsion1, torsion2, weights=weights)
  return tfd


def GetBestTFDBetweenMolecules(mol1, mol2, confId1=-1, useWeights=True, maxDev='equal',
                               symmRadius=2, ignoreColinearBonds=True):
  """ Wrapper to calculate the best TFD between a single conformer of mol1 and all the conformers of mol2
      Important: The two molecules must be isomorphic

      Arguments:
      - mol1:     first instance of the molecule of interest
      - mol2:     second instance the molecule of interest
      - confId1:  conformer index for mol1 (default: first conformer)
      - useWeights: flag for using torsion weights in the TFD calculation
      - maxDev:   maximal deviation used for normalization
                  'equal': all torsions are normalized using 180.0 (default)
                  'spec':  each torsion is normalized using its specific
                           maximal deviation as given in the paper
      - symmRadius: radius used for calculating the atom invariants
                    (default: 2)
      - ignoreColinearBonds: if True (default), single bonds adjacent to
                             triple bonds are ignored
                             if False, alternative not-covalently bound
                             atoms are used to define the torsion

      Return: TFD value
  """
  if (Chem.MolToSmiles(mol1) != Chem.MolToSmiles(mol2)):
    raise ValueError("The two molecules must be instances of the same molecule!")
  mol2 = _getSameAtomOrder(mol1, mol2)
  tl, tlr = CalculateTorsionLists(mol1, maxDev=maxDev, symmRadius=symmRadius,
                                  ignoreColinearBonds=ignoreColinearBonds)
  # first molecule
  torsion1 = CalculateTorsionAngles(mol1, tl, tlr, confId=confId1)
  if useWeights:
    weights = CalculateTorsionWeights(mol1, ignoreColinearBonds=ignoreColinearBonds)
  else:
    weights = None
  best = 1e8
  for conf in mol2.GetConformers():
    # second molecule
    torsion2 = CalculateTorsionAngles(mol2, tl, tlr, confId=conf.GetId())
    tfd = CalculateTFD(torsion1, torsion2, weights=weights)
    best = min(best, tfd)
  return best


def GetTFDMatrix(mol, useWeights=True, maxDev='equal', symmRadius=2, ignoreColinearBonds=True):
  """ Wrapper to calculate the matrix of TFD values for the
      conformers of a molecule.

      Arguments:
      - mol:      the molecule of interest
      - useWeights: flag for using torsion weights in the TFD calculation
      - maxDev:   maximal deviation used for normalization
                  'equal': all torsions are normalized using 180.0 (default)
                  'spec':  each torsion is normalized using its specific
                           maximal deviation as given in the paper
      - symmRadius: radius used for calculating the atom invariants
                    (default: 2)
      - ignoreColinearBonds: if True (default), single bonds adjacent to
                             triple bonds are ignored
                             if False, alternative not-covalently bound
                             atoms are used to define the torsion

      Return: matrix of TFD values
      Note that the returned matrix is symmetrical, i.e. it is the
      lower half of the matrix, e.g. for 5 conformers:
      matrix = [ a,
                 b, c,
                 d, e, f,
                 g, h, i, j]
  """
  tl, tlr = CalculateTorsionLists(mol, maxDev=maxDev, symmRadius=symmRadius,
                                  ignoreColinearBonds=ignoreColinearBonds)
  numconf = mol.GetNumConformers()
  torsions = [
    CalculateTorsionAngles(mol, tl, tlr, confId=conf.GetId()) for conf in mol.GetConformers()
  ]
  tfdmat = []
  if useWeights:
    weights = CalculateTorsionWeights(mol, ignoreColinearBonds=ignoreColinearBonds)
  else:
    weights = None
  for i in range(0, numconf):
    for j in range(0, i):
      tfdmat.append(CalculateTFD(torsions[i], torsions[j], weights=weights))

  return tfdmat
